
#include "stdafx.h"
#include "AstralVw.h"
#include "resource.h"
#include "fprintf.h"
#include "SNESop.h"
#include "AstralFrame.h"

#define _OUTPUT_ZISAZSNES_BRANCH_TXT 1

using namespace Astral;
using namespace SNES;

static CAsselVw *getAsselVw(CWnd *here)
{
	return &static_cast<CAstralFrame *>(here->GetParentFrame())->wndAsselDock.frame.vw;
}

void Astral::collectAstomark(Astomark &r, AstomarkPVec &vec)
{
	vec.clear();

	AstomarkPList ol;
	ol.push_back(&r);
	while (!ol.empty()) {
		Astomark *m = ol.front(); ol.pop_front();
		vec.push_back(m);
		size_t i;
		for (i = 0; i < m->children.size(); i++) {
			ol.push_back(m->children[i]);
		}
	}
}

static CString formatAddr(DWORD addr)
{
	addr &= 0xFFFFFF;
	CString text; text.Format(
		"%02X/%04X"
		, (UINT)(addr >> 16)
		, (UINT)(addr & 0xFFFF)
		);
	return text;
}

// CAstralVw

CAstralVw::CAstralVw()
{
}

CAstralVw::~CAstralVw()
{
}

BEGIN_MESSAGE_MAP(CAstralVw, CScrollView)
	ON_WM_ERASEBKGND()
	ON_COMMAND(ID_FILE_CONVERT, OnFileConvert)
END_MESSAGE_MAP()

void CAstralVw::OnInitialUpdate()
{
	CScrollView::OnInitialUpdate();

	cxbox = 50;
	cybox = 50;

	Reset();
}

void CAstralVw::OnDraw(CDC* pDC)
{
	if (vec.size() == 0) return;

	CDC &dc = *pDC;

	CSize size = GetTotalSize();
	CRect rccl;
	GetClientRect(&rccl);

	bool centerize = size.cx < rccl.Width();
	CPoint renderoffset = centerize
		? CPoint((rccl.Width() - size.cx)/2, 0)
		: CPoint(0, 0)
		;

	CPen pengreen(PS_SOLID, 0, RGB(0, 255, 0));
	CPen pencyan(PS_SOLID, 0, RGB(0, 255, 255));
	CPen penyellow(PS_SOLID, 0, RGB(255, 255, 0));
	CPen penblue(PS_SOLID, 0, RGB(0, 0, 255));
	CPen penred(PS_SOLID, 0, RGB(255, 0, 0));

	dc.SaveDC();
	dc.SelectStockObject(DEFAULT_GUI_FONT);

	size_t i, n = vec.size();
	for (i = 0; i < n; i++) {
		Astomark *mark = vec[i];
		int depth = mark->depth;

		bool childrenok = (mark->children.size() != 0);

		CRect rc = mark->rc + renderoffset;
		CPoint ptc = rc.CenterPoint();

		CRect rccir(ptc, ptc);
		rccir.InflateRect(10, 10, 11, 11);

		dc.SelectStockObject(WHITE_PEN);
		dc.SelectStockObject(BLACK_BRUSH);
		dc.MoveTo(ptc.x, (depth == 0) ? ptc.y : rc.top);
		dc.LineTo(ptc.x, (childrenok) ? rc.bottom : ptc.y);

		if (mark->elt & 0x01) {
			// JSL
			dc.SelectObject(&pencyan);
		}
		else if (mark->elt & 0x04) {
			// JSR
			dc.SelectObject(&pengreen);
		}
		else if (mark->elt & 0x0A) {
			// RTL or RTS
			dc.SelectObject(&penblue);
		}
		else if (mark->elt & 0x10) {
			// branch
			dc.SelectObject(&pencyan);
		}
		else if (mark->elt & 0x40) {
			// branch root
			dc.SelectObject(&pengreen);
		}
		else if (mark->elt & 0x80) {
			// no more
			dc.SelectObject(&penred);
		}
		dc.Ellipse(rccir);

		CString text;
		text.Format("%s", formatAddr(mark->addr));
		dc.SetBkMode(TRANSPARENT);
		dc.SetTextColor(RGB(255,255,0));
		dc.TextOut(rccir.left, rccir.top, text);

		if (childrenok) {
			CRect rc0 = renderoffset + mark->children[0]->rc;
			CRect rc1 = renderoffset + mark->children[mark->children.size() -1]->rc;
			dc.SelectStockObject(WHITE_PEN);
			dc.MoveTo(rc0.CenterPoint().x, rc.bottom);
			dc.LineTo(rc1.CenterPoint().x, rc.bottom);
		}
	}

	dc.RestoreDC(-1);
}

void CAstralVw::Reset()
{
	vec.clear();
	mark.clear();
	xcnt = 0;
	getAsselVw(this)->Reset();

	CSize sizeTotal;
	sizeTotal.cx = sizeTotal.cy = 100;
	SetScrollSizes(MM_TEXT, sizeTotal);
}

BOOL CAstralVw::PreCreateWindow(CREATESTRUCT& cs)
{
	if (!CScrollView::PreCreateWindow(cs)) return false;

	cs.dwExStyle &= ~WS_EX_CLIENTEDGE;
	return true;
}

BOOL CAstralVw::OnEraseBkgnd(CDC* pDC)
{
	CRect rc;
	GetClientRect(&rc);
	pDC->FillSolidRect(rc, RGB(0, 0, 0));
	return true;
}

void CAstralVw::OnFileConvert()
{
	CWaitCursor wc;

	Reset();

	OccurrenceMap occur;
	OccurrenceSet ocsetc;
	ReversedOccurrenceMap revoccur;

	OPENEDList OL;

	{
		WalkAscMap::iterator
			iterPos = walkMapAsc.begin(),
			iterEnd = walkMapAsc.end();
		while (iterPos != iterEnd) {
			{
				BYTE *program = (BYTE *)&iterPos->second.execo;
				Occuro oo;
				oo.addr = iterPos->first.newPos;
				oo.spec = SpecOfOpfaster(program[0]);
				oo.exec = iterPos->second.execo;
				oo.status = iterPos->second.status;
				occur.insert(OccurrenceMap::value_type(iterPos->first.oldPos, oo));

				if (ocsetc.insert(iterPos->first.newPos).second) {
					if (iterPos->first.oldPos == 0xFFFFFFFF) {
						OPENED op;
						op.island.posfirst = op.pos = iterPos->first.newPos;
						OL.push_front(op);
					}
					else if (oo.spec & 0x01) {
						OPENED op;
						op.island.posfirst = op.pos = iterPos->first.newPos;
						op.island.classis = 1;
						OL.push_back(op);
					}
					else if (oo.spec & 0x04) {
						OPENED op;
						op.island.posfirst = op.pos = iterPos->first.newPos;
						op.island.classis = 2;
						OL.push_back(op);
					}
				}
			}

			{
				BYTE *program = (BYTE *)&iterPos->second.execo;
				Occuro oo;
				oo.addr = iterPos->first.oldPos;
				oo.spec = SpecOfOpfaster(program[0]);
				oo.exec = iterPos->second.exec;
				revoccur.insert(ReversedOccurrenceMap::value_type(iterPos->first.newPos, oo));
			}

			iterPos++;
		}
	}

	CLOSEDList CL;

	while (OL.size() != 0) {
		OPENED &op = OL.front();

		size_t n1 = op.occurMe.size();
		size_t n2 = op.evec.size();

		DWORD pos = op.pos;

		OccurrenceSet &ocset = op.occurMe;

		if (!ocset.insert(pos).second) {
			// terminate race: already met
			CL.push_back(op);
			OL.pop_front();
			continue;
		}

		OccurrenceMap::iterator
			iterPos = occur.lower_bound(pos),
			iterEnd = occur.upper_bound(pos);
		size_t n = std::distance(iterPos, iterEnd);

		if (n == 0) {
			// terminate race: no more steps
			CL.push_back(op);
			OL.pop_front();
			continue;
		}
		ASSERT(n < 256);

		char termrace = 0;
		char killrace = 0;
		char continuerace = 0;
		WORD elect = 0;
		BYTE retok = 0;
		BYTE branchfirst = 0;

		while (iterPos != iterEnd) {
			int nJSL = 0, nJSR = 0, nRTL = 0, nRTS = 0, nBranch = 0;

			DWORD posNew = iterPos->second.addr;
			if (posNew & 0x80000000) {
				// nmi/virq interruption call
				iterPos++;
				continue;
			}

			BYTE size2 = SizeOfOp((BYTE)iterPos->second.exec, iterPos->second.status);

			ReversedOccurrenceMap::iterator
				iterrPos = revoccur.lower_bound(pos + size2),
				iterrEnd = revoccur.upper_bound(pos + size2);
			size_t nr = std::distance(iterrPos, iterrEnd);
			while (iterrPos != iterrEnd) {
				BYTE spec = iterrPos->second.spec;
				if (spec & 0x01) nJSL++;
				if (spec & 0x02) nRTL++;
				if (spec & 0x04) nJSR++;
				if (spec & 0x08) nRTS++;
				if (spec & 0x10) nBranch++;
				iterrPos++;
			}

			BYTE spec = iterPos->second.spec;
			if (spec & 0x01) {
				// JSL
				if (nRTL != 0) {
					//op.pos = posNew; continuerace |= 1;

					// new closed result
					killrace |= 1;
					OPENED opNext = op;
					Election el;
					el.pos = pos;
					el.posNew = posNew;
					el.iCur = (BYTE)elect; el.nMax = (255 < n) ? 255 : n; elect++;
					el.nMax = 0;
					el.elt = spec;
					opNext.pos = posNew;
					opNext.evec.push_back(el);

					CL.push_back(opNext);

					iterPos++;
					continue;
				} else {
					////ASSERT(false);
					termrace |= 2;
				}
			}
			else if (spec & 0x04) {
				// JSR
				if (nRTS != 0) {
					//op.pos = posNew; continuerace |= 1;

					// new closed result
					killrace |= 1;
					OPENED opNext = op;
					Election el;
					el.pos = pos;
					el.posNew = posNew;
					el.iCur = (BYTE)elect; el.nMax = (255 < n) ? 255 : n; elect++;
					el.nMax = 0;
					el.elt = spec;
					opNext.pos = posNew;
					opNext.evec.push_back(el);

					CL.push_back(opNext);

					iterPos++;
					continue;
				} else {
					ASSERT(false);
					termrace |= 2;
				}
			}
			else if (spec & 0x0A) {
				// RTL RTS
				termrace |= 1;

				if (spec & 0x02) {
					ASSERT(op.island.classis == 1);
				}
				else if (spec & 0x08) {
					ASSERT(op.island.classis == 2);
				}

				if (retok == 0) {
					retok++;
					// new closed result
					killrace |= 1;

					OPENED opNext = op;

					Election el;
					el.pos = pos;
					el.posNew = posNew;
					el.iCur = (BYTE)elect; el.nMax = (255 < n) ? 255 : n; elect++;
					el.nMax = 0;
					el.elt = spec & 0x0A;

					opNext.pos = posNew;
					opNext.evec.push_back(el);

					CL.push_back(opNext);
				}

				iterPos++;
				continue;
			}
			else if (spec & 0x10) {
				// branch

				killrace |= 1;

				if (ocset.find(posNew) != ocset.end()) {
					// already met!
					iterPos++;
					continue;
				}
				if (elect == 256) {
					// too many branches!
					iterPos++;
					continue;
				}

				if (branchfirst == 0) {
					branchfirst++;

					Election el;
					el.pos = pos;
					el.posNew = posNew;
					el.iCur = 0;
					el.nMax = 0;
					el.elt = 0x40;
					op.evec.push_back(el);
				}

				// branch occurred
				OPENED &opNext = *OL.insert(OL.end(), op);

				Election el;
				el.pos = pos;
				el.posNew = posNew;
				el.iCur = (BYTE)elect; el.nMax = (255 < n) ? 255 : n; elect++;
				el.nMax = 0;
				el.elt = spec;

				opNext.pos = posNew;
				opNext.evec.push_back(el);

				iterPos++;
				continue;
			}
			else {
				// other instructions
				if (pos + size2 == posNew) {
					op.pos = posNew; continuerace |= 1;
				}
				else if (continuerace == 0) {
					////ASSERT(false);
					termrace |= 4;

					// no more execution on this race
					Election el;
					el.pos = pos;
					el.posNew = -1;
					el.iCur = (BYTE)elect; el.nMax = (255 < n) ? 255 : n; elect++;
					el.nMax = 0;
					el.elt = 0x80;
					op.evec.push_back(el);
				}
			}

			iterPos++;
		}

		if (killrace != 0) {
			OL.pop_front();
		}
		else if (termrace != 0 && continuerace == 0) {
			// terminate race: ret detected
			CL.push_back(op);
			OL.pop_front();
		}
	}

	std::stable_sort(CL.begin(), CL.end());

	using namespace std::rel_ops;

	CAsselVw *vw = getAsselVw(this);

	vw->Reset();

	mark.clear();

	Astomark *mark0 = NULL;

	{
		CLOSEDList::iterator
			iterPos = CL.begin(),
			iterEnd = CL.end();
		ISLAND island;
		while (iterPos != iterEnd) {
			CLOSED &r = *iterPos;
			//
			if (r.island != island) {
				island = r.island;
				vw->AddSite(island.posfirst);
				mark0 = mark.create(island.posfirst);
			}
			Astomark *mark1 = mark0;
			ElectionVec &evec = r.evec;
			size_t i;
			for (i = 0; i < evec.size(); i++) {
				const Election &el = evec[i];
				DWORD addr = (el.elt & 0xC0) ? el.pos : el.posNew;
				Astomark *marknew = mark1->create(addr);
				marknew->elt = el.elt;
				mark1 = marknew;
			}
			//
			iterPos++;
		}
	}

	printf("");

#if _OUTPUT_ZISAZSNES_BRANCH_TXT

	{
		CString text;
		ISLAND island;
		while (CL.size() != 0) {
			CLOSED &r = CL.front();
			//
			if (r.island != island) {
				island = r.island;
				text.AppendFormat("\n#%d [%02X/%04X]\n"
					, (UINT)island.classis
					, (UINT)((island.posfirst >>16) & 0x00FF)
					, (UINT)((island.posfirst >> 0) & 0xFFFF)
					);
			}
			CString debugged;
			char sz[1000];
			UINT i = 0;
			while (i < 1000 && i < r.evec.size()) {
				const Election &el = r.evec[i];
			//	if (el.elt != 0) {
			//		i++;
			//		continue;
			//	}
				debugged.AppendFormat("|%02X/%04X"
					, (UINT)((el.pos >> 16) & 0xFF)
					, (UINT)((el.pos >>  0) & 0xFFFF)
					);
				static const char table[] = {"0123456789ABCDEFGHIJKLMNOPQRSUTVWXYZ"};
				UINT x = el.iCur;
				if (x < 36) {
					sz[i] = table[x];
				} else {
					sz[i] = '?';
				}
				i++;
			}
			while (i < 50) {
				sz[i] = ' '; i++;
			}
			sz[i] = 0;
			text += sz; text += debugged; text += '\n';
			//
			CL.pop_front();
		}
		FILE *f = fopen("\\zisazsnes.branch.txt", "wt");
		if (f != NULL) {
			fputs(text, f);
			fclose(f);
		}
	}
#endif
}

void CAstralVw::Select(DWORD addr)
{
	Astomark *mark0 = mark.find(addr);
	if (mark0 == NULL)
		return;

	mark0->depth = 0;
	ycnt = 0;

	collectAstomark(*mark0, vec);

	size_t i, n = vec.size();
	for (i = 0; i < n; i++) {
		Astomark *mark1 = vec[n -i -1];
		mark1->layout0();
	}
	for (i = 0; i < n; i++) {
		Astomark *mark1 = vec[i];
		mark1->layout1();
	}

	xcnt = mark0->xcnt;

	{
		mark0->rc.SetRect(0, 0, cxbox*xcnt, cybox);

		AstomarkPVec::iterator
			iterPos = vec.begin(),
			iterEnd = vec.end();
		int curcx = 0;
		int newcx = xcnt;
		while (iterPos != iterEnd) {
			curcx = newcx;
			newcx = 0;
			Astomark check;
			check.depth = ycnt;
			AstomarkPVec::iterator
				iteroPos = std::lower_bound(iterPos, iterEnd, &check, Astomark::depth_equal()),
				iteroEnd = std::upper_bound(iterPos, iterEnd, &check, Astomark::depth_equal());
			size_t n0 = std::distance(iterPos, iterEnd);
			size_t n1 = std::distance(iteroPos, iteroEnd);
			int y = ycnt;
			while (iteroPos != iteroEnd) {
				int x = 0;
				Astomark *mark0 = *iteroPos;
				CRect rc0 = mark0->rc;
				size_t i, n = mark0->children.size();
				for (i = 0; i < n; i++) {
					Astomark *mark1 = mark0->children[i];
					int cxmark1 = mark1->xcnt;
					mark1->rc.SetRect(
						rc0.left + (x +0      ) * cxbox,
						rc0.top  + 1            * cybox,
						rc0.left + (x +cxmark1) * cxbox,
						rc0.top  + 2            * cybox
						);
					x += cxmark1;
				}
				newcx += x;
				iteroPos++;
			}
			iterPos = iteroEnd;
			ycnt++;
		}
	}

	Layout();
}

void CAstralVw::Layout()
{
	CSize sizeTotal;
	sizeTotal.cx = cxbox * (xcnt);
	sizeTotal.cy = cybox * (ycnt+1);
	SetScrollSizes(MM_TEXT, sizeTotal);
	Invalidate();
}
