
#pragma once

namespace Astral {
	// 
	typedef std::set<DWORD> OccurrenceSet;
	// 
	struct Election {
		DWORD pos, posNew;
		BYTE iCur, nMax;
		BYTE elt; // 0=BRANCH 1=CALL
	};
	// 
	typedef std::vector<Election> ElectionVec;
	// 
	struct ISLAND {
		DWORD posfirst;
		char classis; // 0=BASE(00/8000) 1=LARGER(JSL) 2=SMALLER(JSR)

		ISLAND(): posfirst(0), classis(0) { }

		inline bool operator <(const ISLAND &r) const {
			if (classis < r.classis) return true;
			if (classis > r.classis) return false;
			if (posfirst < r.posfirst) return true;
			if (posfirst > r.posfirst) return false;
			return false;
		}
		inline bool operator ==(const ISLAND &r) const {
			return classis == r.classis && posfirst == r.posfirst;
		}
	};
	// 
	struct OPENED {
		DWORD pos;
		ISLAND island;
		ElectionVec evec;
		OccurrenceSet occurMe;
	};
	// 
	typedef std::list<OPENED> OPENEDList;

	// 
	struct CLOSED {
		DWORD pos, posNew;
		ISLAND island;
		ElectionVec evec;

		CLOSED() { }
		CLOSED(const OPENED &op)
			: pos(op.pos), evec(op.evec), island(op.island)
		{

		}

		inline bool operator <(const CLOSED &r) const {
			return island < r.island;
		}
	};
	// 
	typedef std::list<CLOSED> CLOSEDList;

	// 
	struct Occuro {
		DWORD addr;
		DWORD exec;
		BYTE spec; // 1=JSL 2=RTL 4=JSR 8=RTS 16=Branch
		BYTE status;
	};
	// 
	typedef std::multimap<DWORD, Occuro> OccurrenceMap;
	typedef std::multimap<DWORD, Occuro> ReversedOccurrenceMap;

	// 
	struct Astomark {
		struct depth_equal {
			bool operator ()(const Astomark *s, const Astomark *t) const {
				return (s->depth < t->depth);
			}
		};

		BYTE elt;
		DWORD addr;
		std::vector<Astomark *> children;

		int xcnt;
		int depth;
		CRect rc;

		Astomark() {
			elt = 0;
			addr = -1;
			xcnt = -1;
			depth = -1;
		}
		~Astomark() {
			clear();
		}
		void clear() {
			size_t i;
			for (i = 0; i < children.size(); i++) delete children[i];
			children.clear();
		}
		Astomark *find(DWORD target) {
			size_t i;
			for (i = 0; i < children.size(); i++) {
				Astomark *child = children[i];
				if (child->addr != target)
					continue;
				return children[i];
			}
			return NULL;
		}
		Astomark *create(DWORD target) {
			Astomark *mark = find(target);
			if (mark != NULL) return mark;
			mark = new Astomark();
			mark->addr = target;
			children.push_back(mark);
			return mark;
		}

		void layout0() {
			size_t i, n = children.size();
			xcnt = 0;
			for (i = 0; i < n; i++) {
				xcnt += children[i]->xcnt;
			}
			if (xcnt == 0) {
				xcnt = 1;
			}
			return;
		}
		void layout1() {
			size_t i, n = children.size();
			for (i = 0; i < n; i++) {
				children[i]->depth = depth + 1;
			}
		}
	};
	// 
	typedef std::vector<Astomark *> AstomarkPVec;
	typedef std::list<Astomark *> AstomarkPList;
	// 
	void collectAstomark(Astomark &r, AstomarkPVec &vec);

};

class CAstralVw : public CScrollView
{
	// 
	DECLARE_MESSAGE_MAP()

public:
	// 
	Astral::Astomark mark;

	int xcnt, ycnt;
	int cxbox, cybox;
	Astral::AstomarkPVec vec;

	// 
	CAstralVw();
	virtual ~CAstralVw();
	virtual void PostNcDestroy() { }
	// 
	void Reset();
	void Select(DWORD addr);
	void Layout();

protected:
	// 
	virtual void OnDraw(CDC* pDC);
	virtual void OnInitialUpdate();

	virtual BOOL PreCreateWindow(CREATESTRUCT& cs);
public:
	afx_msg BOOL OnEraseBkgnd(CDC* pDC);
	afx_msg void OnFileConvert();
};
