
#include "stdafx.h"
#include "CustomVw.h"
#include "resource.h"
#include "ScriptoDlg.h"
#include "fprintf.h"
#include "ScriptStatDlg.h"
#include "CustomFrame.h"
#include "IDEFrame.h"

extern CIDEFrame frameide;

// CCustomVw

CCustomVw::CCustomVw()
{
	iNextNumber = 0;
	fAlwaysTop = true;
	fCapture = true;
}

CCustomVw::~CCustomVw()
{

}

BEGIN_MESSAGE_MAP(CCustomVw, CListView)
	ON_WM_CREATE()
	ON_COMMAND(ID_POPUP_SCRIPTO, OnPopupScripto)
	ON_COMMAND(ID_BUTTON_CLEAR, OnButtonClear)
	ON_COMMAND(ID_BUTTON_ALWAYS_TOP, OnButtonAlwaysTop)
	ON_UPDATE_COMMAND_UI(ID_BUTTON_ALWAYS_TOP, OnUpdateButtonAlwaysTop)
	ON_COMMAND(ID_BUTTON_PLAY, OnButtonPlay)
	ON_UPDATE_COMMAND_UI(ID_BUTTON_PLAY, OnUpdateButtonPlay)
	ON_COMMAND(ID_POPUP_SCRIPT_STAT, OnPopupScriptStat)
	ON_NOTIFY_REFLECT(LVN_DELETEITEM, OnLvnDeleteitem)
	ON_NOTIFY_REFLECT(LVN_ITEMACTIVATE, OnLvnItemActivate)
	ON_NOTIFY_REFLECT(LVN_ITEMCHANGED, OnLvnItemchanged)
END_MESSAGE_MAP()

#define LC GetListCtrl()

void CCustomVw::OnInitialUpdate()
{
	CListView::OnInitialUpdate();

	while (LC.DeleteColumn(0));

	LC.DeleteAllItems();

	LC.SetExtendedStyle(LC.GetExtendedStyle() |LVS_EX_FULLROWSELECT);

	LC.SetBkColor(RGB(0, 0, 88));
	LC.SetTextColor(RGB(255, 255, 255));
	LC.SetTextBkColor(RGB(0, 0, 88));

#define LCC_SRC			0
#define LCC_DATETIME	1
#define LCC_NUMBER		2
#define LCC_MES			3

	LC.InsertColumn(LCC_SRC, "Sev/Src", LVCFMT_LEFT, 80);
	LC.InsertColumn(LCC_DATETIME, "DateTime", LVCFMT_LEFT, 1);
	LC.InsertColumn(LCC_NUMBER, "Number", LVCFMT_LEFT, 50);
	LC.InsertColumn(LCC_MES, "Message                       ", LVCFMT_LEFT, 300);

	LC.SetFont(&m_fontLv);

	LC.SetImageList(&m_photo, LVSIL_SMALL);

	LC.SetColumnWidth(LCC_MES, LVSCW_AUTOSIZE_USEHEADER);

//	LC.InsertItem(0, "DEFAULT", 0);
//	LC.InsertItem(1, "INFO", 1);
//	LC.InsertItem(2, "WARN", 2);
//	LC.InsertItem(3, "ERR", 3);
//	LC.InsertItem(4, "X", 4);
}

BOOL CCustomVw::PreCreateWindow(CREATESTRUCT& cs)
{
	if (!CListView::PreCreateWindow(cs)) return false;

	cs.style &= ~LVS_TYPEMASK;
	cs.style |= 0
		| LVS_REPORT
		| LVS_SINGLESEL
		| LVS_SHOWSELALWAYS
		| LVS_SHAREIMAGELISTS
		| LVS_NOSORTHEADER
		;
	return true;
}

int CCustomVw::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
	if (CListView::OnCreate(lpCreateStruct) == -1)
		return -1;

	LOGFONT lf;
	GetObject(GetStockObject(DEFAULT_GUI_FONT), sizeof(LOGFONT), &lf);
	_tcscpy(lf.lfFaceName, _T(""));
	lf.lfPitchAndFamily = 0 |FIXED_PITCH;

	if (!m_fontLv.CreateFontIndirect(&lf))
		return -1;

	const int cx = GetSystemMetrics(SM_CXSMICON);
	const int cy = GetSystemMetrics(SM_CYSMICON);

	m_photo.DeleteImageList();
	if (!m_photo.Create(cx, cy, 0 |ILC_MASK |ILC_COLORDDB, 0, 1))
		return -1;

	m_photo.Add(AfxGetApp()->LoadIcon(IDR_CUST));
	m_photo.Add(AfxGetApp()->LoadOEMIcon((UINT)IDI_INFORMATION));
	m_photo.Add(AfxGetApp()->LoadOEMIcon((UINT)IDI_WARNING));
	m_photo.Add(AfxGetApp()->LoadOEMIcon((UINT)IDI_ERROR));

	ShowScrollBar(SB_VERT);

	return 0;
}

void CCustomVw::OnPopupScripto()
{
	ScriptoVec bkup = scriptoVec;

	CScriptoDlg wndDlg(AfxGetMainWnd());
	if (wndDlg.DoModal() != IDOK) {
		scriptoVec = bkup;
		return;
	}
}

namespace {

CString FormatDateTime_Now()
{
	CTime t(mktime(CTime::GetCurrentTime().GetLocalTm()));
	return t.Format(_T("%Y-%m-%d %H:%M:%S"));
}

};

void CCustomVw::Message(LPCTSTR pszMessage, LPCSTR pszSrc, int severity, int &refid)
{
	refid = -1;

	if (!fCapture) return;

	static LPCTSTR pszSevere[] = {
		"----",
		"INFO",
		"WARN",
		"ERR!",
	};
	severity &= 3;

	int r = LC.InsertItem(LC.GetItemCount(), pszSrc, severity);
	if (r < 0) return;

	Data *ref = new Data();
	LC.SetItemData(r, reinterpret_cast<DWORD_PTR>(ref));

	refid = reinterpret_cast<int>(ref);

	static char szNumber[100];
	sprintf(szNumber, "%6u", iNextNumber);
	iNextNumber++;

	int n = _tcslen(pszMessage);
	LPCTSTR pszMark = _tcschr(pszMessage, '\n');
	if (pszMark != NULL) {
		n = pszMark - pszMessage;
	}

	ref->strDescription = pszMessage;

	LC.SetItemText(r, LCC_DATETIME, FormatDateTime_Now());
	LC.SetItemText(r, LCC_NUMBER, szNumber);
	LC.SetItemText(r, LCC_MES, ref->strDescription.Left(n));

	if (fAlwaysTop) {
		LC.EnsureVisible(r, false);
	}
}

CCustReport *CCustomVw::FindReport(int refid, bool invalidate)
{
	LVFINDINFO lvfi;
	ZeroMemory(&lvfi, sizeof(lvfi));

	lvfi.flags = LVFI_PARAM;
	lvfi.lParam = (LPARAM)refid;

	int r = LC.FindItem(&lvfi);
	if (r != -1) {
		if (invalidate) {
			int rr = LC.GetNextItem(-1, LVNI_SELECTED);
			if (rr == r) {
				SelEntry(r);
			}
		}
		return reinterpret_cast<CCustReport *>(refid);
	}
	return NULL;
}

void CCustomVw::Reset()
{
	iNextNumber = 0;

//	static_cast<CCustomFrame *>(GetParentFrame())->wndFRDock.Reset();

	frameide.wndFRDock.Reset();

	LC.DeleteAllItems();
}

void CCustomVw::OnButtonClear()
{
	Reset();
}

void CCustomVw::OnButtonAlwaysTop()
{
	fAlwaysTop = !fAlwaysTop;
}

void CCustomVw::OnUpdateButtonAlwaysTop(CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck(fAlwaysTop);
}

void CCustomVw::OnButtonPlay()
{
	fCapture = !fCapture;
}

void CCustomVw::OnUpdateButtonPlay(CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck(fCapture);
}

void CCustomVw::OnPopupScriptStat()
{
	CScriptStatDlg wndDlg(AfxGetMainWnd());
	wndDlg.DoModal();
}

void CCustomVw::OnLvnDeleteitem(NMHDR *pNMHDR, LRESULT *pResult)
{
	LPNMLISTVIEW pNMLV = reinterpret_cast<LPNMLISTVIEW>(pNMHDR);
	//
	Data *ref = reinterpret_cast<Data *>(pNMLV->lParam);
	if (ref != NULL) delete ref;
	//
	*pResult = 0;
}

void CCustomVw::OnLvnItemActivate(NMHDR *pNMHDR, LRESULT *pResult)
{
	LPNMITEMACTIVATE pNMIA = reinterpret_cast<LPNMITEMACTIVATE>(pNMHDR);
	//
	//
	*pResult = 0;
}

void CCustomVw::OnLvnItemchanged(NMHDR *pNMHDR, LRESULT *pResult)
{
	LPNMLISTVIEW pNMLV = reinterpret_cast<LPNMLISTVIEW>(pNMHDR);
	//
	if (pNMLV->uNewState & LVIS_SELECTED) {
		SelEntry(pNMLV->iItem);
	}
	//
	*pResult = 0;
}

void CCustomVw::SelEntry(int i)
{
	Data *ref = reinterpret_cast<Data *>(LC.GetItemData(i));
	CString strReport;
	if (ref != NULL) {
		strReport = ref->strDescription;
	}
	frameide.wndFRDock.SetReport(strReport, ref->report);
}
