
#include "stdafx.h"
#include "DebVw.h"
#include "resource.h"
#include "fprintf.h"
#include "SNESop.h"
#include "fprintfi.h"

#define LC GetListCtrl()

// FOR OPERAND FORMAT

namespace {

inline void printhex4(char *psz, BYTE x)
{
	static const BYTE table[] = {"0123456789ABCDEF"};
	*psz = table[(x     ) & 15];
}

inline void printhex8(char *psz, BYTE x)
{
	static const BYTE table[] = {"0123456789ABCDEF"};
	*psz = table[(x >> 4)     ]; psz++;
	*psz = table[(x     ) & 15];
}

inline void printhex16(char *psz, WORD x)
{
	static const BYTE table[] = {"0123456789ABCDEF"};
	*psz = table[(x >>12)     ]; psz++;
	*psz = table[(x >> 8) & 15]; psz++;
	*psz = table[(x >> 4) & 15]; psz++;
	*psz = table[(x     ) & 15];
}

#define writechar(A)	sz[x] = (BYTE)(A), x++;
#define	writehex4(A)	printhex4(&sz[x], (BYTE)(A)), x++;
#define	writehex8(A)	printhex8(&sz[x], (BYTE)(A)), x += 2;
#define	writehex16(A)	printhex16(&sz[x], (WORD)(A)), x += 4;

#define READBYTE(P)		(*(const BYTE *)(P))
#define READWORD(P)		(*(const WORD *)(P))

};

// CDebVw

CDebVw::CDebVw()
{

}

CDebVw::~CDebVw()
{

}

BEGIN_MESSAGE_MAP(CDebVw, CListView)
	ON_WM_CREATE()
	ON_COMMAND(ID_FILE_CONVERT, OnFileConvert)
	ON_NOTIFY_REFLECT(LVN_GETDISPINFO, OnLvnGetdispinfo)
	ON_COMMAND(ID_FILE_SAVE_COPY_AS, OnFileSaveCopyAs)
	ON_NOTIFY_REFLECT(LVN_ODFINDITEM, OnLvnOdfinditem)
	ON_COMMAND(ID_EDIT_COPY, OnEditCopy)
	ON_COMMAND(ID_FILE_GRAF, OnFileGraf)

	ON_COMMAND_RANGE(ID_BUTTON_CPU, ID_BUTTON_SA1, OnButtonSa1)
	ON_UPDATE_COMMAND_UI_RANGE(ID_BUTTON_CPU, ID_BUTTON_SA1, OnUpdateButtonSa1)
END_MESSAGE_MAP()

#ifdef _DEBUG
void CDebVw::AssertValid() const
{
	CListView::AssertValid();
}

void CDebVw::Dump(CDumpContext& dc) const
{
	CListView::Dump(dc);
}
#endif //_DEBUG

void CDebVw::OnInitialUpdate()
{
	CListView::OnInitialUpdate();

	prot = 0;

	LC.SetExtendedStyle(LC.GetExtendedStyle() |LVS_EX_FULLROWSELECT);

	LC.SetBkColor(RGB(0, 0, 88));
	LC.SetTextColor(RGB(255, 255, 255));
	LC.SetTextBkColor(RGB(0, 0, 88));

	while (LC.DeleteColumn(0));

#define LCC_PBPC 0
#define LCC_COUNT 1
#define LCC_BC 2
#define LCC_INST 3

	LC.InsertColumn(0, "PB/PC", LVCFMT_LEFT, 80);
	LC.InsertColumn(1, "Occur.", LVCFMT_RIGHT, 60);
	LC.InsertColumn(2, "B.C.", LVCFMT_LEFT, 100);
	LC.InsertColumn(3, "Assembler", LVCFMT_LEFT, 300);

	LC.SetFont(&m_fontLv);
}

BOOL CDebVw::PreCreateWindow(CREATESTRUCT& cs)
{
	if (!CListView::PreCreateWindow(cs)) return false;

	cs.style &= ~LVS_TYPEMASK;
	cs.style |= 0
		| LVS_REPORT
		| LVS_SINGLESEL
		| LVS_SHOWSELALWAYS
		| LVS_SHAREIMAGELISTS
		| LVS_OWNERDATA
//		| LVS_OWNERDRAWFIXED
		| LVS_NOSORTHEADER
		;
	return true;
}

int CDebVw::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
	if (CListView::OnCreate(lpCreateStruct) == -1)
		return -1;

	LOGFONT lf;
	GetObject(GetStockObject(DEFAULT_GUI_FONT), sizeof(LOGFONT), &lf);
	_tcscpy(lf.lfFaceName, _T(""));
	lf.lfPitchAndFamily = 0 |FIXED_PITCH;

	if (!m_fontLv.CreateFontIndirect(&lf))
		return -1;

	return 0;
}

void CDebVw::UseProt1()
{
	Line oldLine;
	{
		UINT index = LC.GetTopIndex();
		if (index < lineVec.GetCount()) {
			oldLine = lineVec[index];
		}
	}

	LC.SetItemCount(0);

	Addr2LabelMap labelMap;
	AddrRefMap addrMap;

	{
		AddrBranchMap branchMap;

		{
			WalkAscMap::iterator
				iterPos = walkMapAPU.begin(),
				iterEnd = walkMapAPU.end();
			for (; iterPos != iterEnd; iterPos++) {
				const WalkAsc &asc = iterPos->first;
				const WalkRec &rec = iterPos->second;

				if (asc.oldPos == -1) continue;

				BYTE op = (BYTE)rec.execo;
				if (SNES::instapu[op].branch & 1) {
					addrMap[asc.oldPos];
					branchMap.insert(std::make_pair(asc.oldPos, asc.newPos));
				}
			}
		}

		{
			AddrRefMap::iterator
				iterPos = addrMap.begin(),
				iterEnd = addrMap.end();
			for (DWORD index = 0; iterPos != iterEnd; iterPos++, index++) {
				AddrBranchMap::iterator
					iterBPos = branchMap.lower_bound(iterPos->first),
					iterBEnd = branchMap.upper_bound(iterPos->first);
				DWORD branch = 0;
				for (; iterBPos != iterBEnd; iterBPos++, branch++) {
					Label v;
					v.no = index;
					v.branch = branch;

					labelMap.insert(std::make_pair(iterBPos->second, v));
				}
				iterPos->second.no = index;
				iterPos->second.n = branch;
			}
		}
	}

	lineVec.SetCount(0, walkMapAPU.size() + labelMap.size());

	DWORD oldPos = -1;
	WalkAscMap::iterator
		iterPos = walkMapAPU.begin(),
		iterEnd = walkMapAPU.end();
	for (; iterPos != iterEnd; iterPos++) {
		const WalkAsc &asc = iterPos->first;
		const WalkRec &rec = iterPos->second;

		if (asc.oldPos == oldPos) {
			Line &r = lineVec[lineVec.GetCount() - 1];
			r.occur += rec.occur;
			continue;
		}
		if (asc.oldPos == -1) continue;

		oldPos = asc.oldPos;

		// http://snescentral.edgeemu.com/Development/Tutorials/01asm.htm

		{
			Addr2LabelMap::iterator
				iterPos = labelMap.lower_bound(oldPos),
				iterEnd = labelMap.upper_bound(oldPos);
			for (; iterPos != iterEnd; iterPos++) {
				Line r;
				r.label = true;
				r.labelx = iterPos->second;
				r.offset = oldPos;

				lineVec.Add(r);
			}
		}

		Line r;

		{
			AddrRefMap::iterator
				iterPos = addrMap.find(asc.oldPos),
				iterEnd = addrMap.end();
			if (iterPos != iterEnd) {
				r.ref = true;
				r.refx = iterPos->second;
			}
		}

		r.occur = rec.occur;
		r.offset = asc.oldPos;
		r.exec = rec.exec;
		r.execo = rec.execo;
		r.status = rec.status;

		lineVec.Add(r);
	}

	LC.SetItemCount(lineVec.GetCount());

	{
		UINT i = 0;
		for (; i < lineVec.GetCount(); i++) {
			Line &r = lineVec[i];
			if (false
				|| (oldLine.label && oldLine.labelx == r.labelx && oldLine.offset == r.offset)
				|| (!lineVec[i].label && lineVec[i].offset == oldLine.offset)
			) {
				CRect rc0, rc1;
				if (true
					&&  LC.GetItemRect(0, rc0, LVIR_BOUNDS)
					&&  LC.GetItemRect(i, rc1, LVIR_BOUNDS)
				) {
					LC.Scroll(CSize(0, rc1.top - rc0.top));
					LC.SetItemState(i, 
						0 |LVIS_FOCUSED |LVIS_SELECTED, 
						0 |LVIS_FOCUSED |LVIS_SELECTED
						);
				}
				break;
			}
		}
	}
}

void CDebVw::UseProt2()
{
	Line oldLine;
	{
		UINT index = LC.GetTopIndex();
		if (index < lineVec.GetCount()) {
			oldLine = lineVec[index];
		}
	}

	LC.SetItemCount(0);

	Addr2LabelMap labelMap;
	AddrRefMap addrMap;

	{
		AddrBranchMap branchMap;

		{
			WalkAscMap::iterator
				iterPos = walkMapSA1.begin(),
				iterEnd = walkMapSA1.end();
			for (; iterPos != iterEnd; iterPos++) {
				const WalkAsc &asc = iterPos->first;
				const WalkRec &rec = iterPos->second;

				if (asc.oldPos == -1 || asc.newPos & 0x80000000) continue;

				BYTE op = (BYTE)rec.execo;
				if (SNES::instruc[op].branch & 1) {
					addrMap[asc.oldPos];
					branchMap.insert(std::make_pair(asc.oldPos, asc.newPos));
				}
			}
		}

		{
			AddrRefMap::iterator
				iterPos = addrMap.begin(),
				iterEnd = addrMap.end();
			for (DWORD index = 0; iterPos != iterEnd; iterPos++, index++) {
				AddrBranchMap::iterator
					iterBPos = branchMap.lower_bound(iterPos->first),
					iterBEnd = branchMap.upper_bound(iterPos->first);
				DWORD branch = 0;
				for (; iterBPos != iterBEnd; iterBPos++, branch++) {
					Label v;
					v.no = index;
					v.branch = branch;

					labelMap.insert(std::make_pair(iterBPos->second, v));
				}
				iterPos->second.no = index;
				iterPos->second.n = branch;
			}
		}
	}

	lineVec.SetCount(0, walkMapSA1.size() + labelMap.size());

	DWORD oldPos = -1;
	WalkAscMap::iterator
		iterPos = walkMapSA1.begin(),
		iterEnd = walkMapSA1.end();
	for (; iterPos != iterEnd; iterPos++) {
		const WalkAsc &asc = iterPos->first;
		const WalkRec &rec = iterPos->second;

		if (asc.oldPos == oldPos && lineVec.GetCount() != 0) {
			Line &r = lineVec[lineVec.GetCount() - 1];
			r.occur += rec.occur;
			continue;
		}
		if (asc.oldPos == -1) continue;

		oldPos = asc.oldPos;

		// http://snescentral.edgeemu.com/Development/Tutorials/01asm.htm

		{
			Addr2LabelMap::iterator
				iterPos = labelMap.lower_bound(oldPos),
				iterEnd = labelMap.upper_bound(oldPos);
			for (; iterPos != iterEnd; iterPos++) {
				Line r;
				r.label = true;
				r.labelx = iterPos->second;
				r.offset = oldPos;

				lineVec.Add(r);
			}
		}

		Line r;

		{
			AddrRefMap::iterator
				iterPos = addrMap.find(asc.oldPos),
				iterEnd = addrMap.end();
			if (iterPos != iterEnd) {
				r.ref = true;
				r.refx = iterPos->second;
			}
		}

		r.occur = rec.occur;
		r.offset = asc.oldPos;
		r.exec = rec.exec;
		r.execo = rec.execo;
		r.status = rec.status;

		lineVec.Add(r);
	}

	LC.SetItemCount(lineVec.GetCount());

	{
		UINT i = 0;
		for (; i < lineVec.GetCount(); i++) {
			Line &r = lineVec[i];
			if (false
				|| (oldLine.label && oldLine.labelx == r.labelx && oldLine.offset == r.offset)
				|| (!lineVec[i].label && lineVec[i].offset == oldLine.offset)
			) {
				CRect rc0, rc1;
				if (true
					&&  LC.GetItemRect(0, rc0, LVIR_BOUNDS)
					&&  LC.GetItemRect(i, rc1, LVIR_BOUNDS)
				) {
					LC.Scroll(CSize(0, rc1.top - rc0.top));
					LC.SetItemState(i, 
						0 |LVIS_FOCUSED |LVIS_SELECTED, 
						0 |LVIS_FOCUSED |LVIS_SELECTED
						);
				}
				break;
			}
		}
	}
}

void CDebVw::UseProt0()
{
	Line oldLine;
	{
		UINT index = LC.GetTopIndex();
		if (index < lineVec.GetCount()) {
			oldLine = lineVec[index];
		}
	}

	LC.SetItemCount(0);

	Addr2LabelMap labelMap;
	AddrRefMap addrMap;

	{
		AddrBranchMap branchMap;

		{
			WalkAscMap::iterator
				iterPos = walkMapAsc.begin(),
				iterEnd = walkMapAsc.end();
			for (; iterPos != iterEnd; iterPos++) {
				const WalkAsc &asc = iterPos->first;
				const WalkRec &rec = iterPos->second;

				if (asc.oldPos == -1 || asc.newPos & 0x80000000) continue;

				BYTE op = (BYTE)rec.execo;
				if (SNES::instruc[op].branch & 1) {
					addrMap[asc.oldPos];
					branchMap.insert(std::make_pair(asc.oldPos, asc.newPos));
				}
			}
		}

		{
			AddrRefMap::iterator
				iterPos = addrMap.begin(),
				iterEnd = addrMap.end();
			for (DWORD index = 0; iterPos != iterEnd; iterPos++, index++) {
				AddrBranchMap::iterator
					iterBPos = branchMap.lower_bound(iterPos->first),
					iterBEnd = branchMap.upper_bound(iterPos->first);
				DWORD branch = 0;
				for (; iterBPos != iterBEnd; iterBPos++, branch++) {
					Label v;
					v.no = index;
					v.branch = branch;

					labelMap.insert(std::make_pair(iterBPos->second, v));
				}
				iterPos->second.no = index;
				iterPos->second.n = branch;
			}
		}
	}

	lineVec.SetCount(0, walkMapAsc.size() + labelMap.size());

	DWORD oldPos = -1;
	WalkAscMap::iterator
		iterPos = walkMapAsc.begin(),
		iterEnd = walkMapAsc.end();
	for (; iterPos != iterEnd; iterPos++) {
		const WalkAsc &asc = iterPos->first;
		const WalkRec &rec = iterPos->second;

		if (asc.oldPos == oldPos) {
			Line &r = lineVec[lineVec.GetCount() - 1];
			r.occur += rec.occur;
			continue;
		}
		if (asc.oldPos == -1) continue;

		oldPos = asc.oldPos;

		// http://snescentral.edgeemu.com/Development/Tutorials/01asm.htm

		{
			Addr2LabelMap::iterator
				iterPos = labelMap.lower_bound(oldPos),
				iterEnd = labelMap.upper_bound(oldPos);
			for (; iterPos != iterEnd; iterPos++) {
				Line r;
				r.label = true;
				r.labelx = iterPos->second;
				r.offset = oldPos;

				lineVec.Add(r);
			}
		}

		Line r;

		{
			AddrRefMap::iterator
				iterPos = addrMap.find(asc.oldPos),
				iterEnd = addrMap.end();
			if (iterPos != iterEnd) {
				r.ref = true;
				r.refx = iterPos->second;
			}
		}

		r.occur = rec.occur;
		r.offset = asc.oldPos;
		r.exec = rec.exec;
		r.execo = rec.execo;
		r.status = rec.status;

		lineVec.Add(r);
	}

	LC.SetItemCount(lineVec.GetCount());

	{
		UINT i = 0;
		for (; i < lineVec.GetCount(); i++) {
			Line &r = lineVec[i];
			if (false
				|| (oldLine.label && oldLine.labelx == r.labelx && oldLine.offset == r.offset)
				|| (!lineVec[i].label && lineVec[i].offset == oldLine.offset)
			) {
				CRect rc0, rc1;
				if (true
					&&  LC.GetItemRect(0, rc0, LVIR_BOUNDS)
					&&  LC.GetItemRect(i, rc1, LVIR_BOUNDS)
				) {
					LC.Scroll(CSize(0, rc1.top - rc0.top));
					LC.SetItemState(i, 
						0 |LVIS_FOCUSED |LVIS_SELECTED, 
						0 |LVIS_FOCUSED |LVIS_SELECTED
						);
				}
				break;
			}
		}
	}
}

void CDebVw::OnFileConvert()
{
	CWaitCursor wc;

	if (prot == 1)
		return UseProt1();
	else if (prot == 2)
		return UseProt2();
	else
		return UseProt0();
}

using namespace SNES;

extern "C" BYTE ocname[];
extern "C" BYTE addrmode[];
extern "C" BYTE spcnametab[8*256];
extern "C" BYTE AddressTable[16*256];
extern "C" BYTE ArgumentTable[256];

void CDebVw::OnLvnGetdispinfo(NMHDR *pNMHDR, LRESULT *pResult)
{
	NMLVDISPINFO *pDispInfo = reinterpret_cast<NMLVDISPINFO*>(pNMHDR);
	//
	LVITEM &lvi = pDispInfo->item;
	UINT index = lvi.iItem;
	if (index < lineVec.GetCount()) {
		if (lvi.mask & LVIF_TEXT) {
			strDelay = GetHintAt(index, lvi.iSubItem);
			lvi.pszText = const_cast<LPTSTR>((LPCTSTR)strDelay);
		}
	}
	//
	*pResult = 0;
}

void CDebVw::ResetContents()
{
	lineVec.RemoveAll();

	if (m_hWnd == NULL) return;

	LC.SetItemCount(0);
}

void CDebVw::OnFileSaveCopyAs()
{
	CWaitCursor wc;

	CString str;
	{
		for (UINT i = 0; i < lineVec.GetCount(); i++) {
			CString strColumns[] = {
				GetHintAt(i, 0),
				GetHintAt(i, 1),
				GetHintAt(i, 2),
				GetHintAt(i, 3),
			};
			CString strLine;
			strLine.Format("|%7s|%10s|%-15s|%s\n"
				, (LPCTSTR)strColumns[0]
				, (LPCTSTR)strColumns[1]
				, (LPCTSTR)strColumns[2]
				, (LPCTSTR)strColumns[3]
				);
			str += strLine;
		}
	}

	CFileDialog wndDlg(
		false,
		"txt",
		"fprintf.txt",
		nDefSaveFlags,
		"txt (*.txt)|*.txt|all (*.*)|*.*||",
		this
		);
	if (wndDlg.DoModal() != IDOK) return;
	FILE *f = fopen(wndDlg.GetPathName(), "wt");
	if (f != NULL) {
		if (!(fputs(str, f) < 0)) {
			fclose(f);
			return;
		}
		fclose(f);
	}
	AfxMessageBox("ERR", 0 |MB_ICONEXCLAMATION);
}

CString CDebVw::GetHintProt2(UINT row, UINT col)
{
	const UINT index = row;

	CString strDelay;

	if (!(index < lineVec.GetCount())) return "?";
	Line &r = lineVec[index];

	switch (col) {
	case LCC_PBPC:
		if (r.label) break;

		strDelay.Format("%02X/%04X"
			, 0U +((r.offset & 0xFF0000) >> 16)
			, 0U +((r.offset & 0x00FFFF))
			);
		break;
	case LCC_COUNT:
		if (r.label) break;

		strDelay.Format("%ld", 0L +r.occur);
		break;
	case LCC_BC:
		{
			if (r.label) break;

			char sz[100] = {""};
			BYTE *program = (BYTE *)&r.execo;
			BYTE n = SizeOfOp(program[0], r.status);
			for (BYTE i = 0; i < n; i++) {
				sprintf(strchr(sz, 0), "%02X ", 0U +program[i]);
			}
			strDelay = sz;
			break;
		}
	case LCC_INST:
		{
			if (r.label) {
				strDelay.Format("L%d.%d:"
					, 1U +r.labelx.no
					, 1U +r.labelx.branch
					);
				break;
			}

			BYTE v = (BYTE)r.execo;
			strDelay = CString((LPCSTR)&ocname[4*v], 4);
			BYTE *program = (BYTE *)&r.execo;

			char sz[100] = {""};
			char x = 0;
			switch (addrmode[v]) { // technique and knowledge from debug.asm
			case 0:
				break;
			case 1: // ; #$12,#$1234 (M-flag)
				writechar('#');
				writechar('$');
				if (r.status & 0x20) {
					// M-flag = 1 ---> 8-bit access
					writehex8(READBYTE(&program[1]));
					break;
				}
				// M-flag = 0 ---> 16-bit access
				writehex16(READWORD(&program[1]));
				break;
			case 2: // ; $1234 : db+$1234
				writechar('$');
				writehex16(READWORD(&program[1]));
				break;
			case 3: // ; $123456
				writechar('$');
				writehex8(READBYTE(&program[3]));
				writehex16(READWORD(&program[1]));
				break;
			case 4: // ; $12 : $12+d
				writechar('$');
				writehex8(READBYTE(&program[1]));
				break;
			case 5: // ; A
				writechar('A');
				break;
			case 6: // ; i
				break;
			case 7: // ; ($12),y : ($12+$13+d)+y
				writechar('(');
				writechar('$');
				writehex8(READBYTE(&program[1]));
				writechar(')');
				writechar(',');
				writechar('y');
				break;
			case 8: // ; [$12],y : [$12+$13+$14+d]+y
				writechar('[');
				writechar('$');
				writehex8(READBYTE(&program[1]));
				writechar(']');
				writechar(',');
				writechar('y');
				break;
			case 9: // ; ($12,x) : ($12+$13+d+x)
				writechar('(');
				writechar('$');
				writehex8(READBYTE(&program[1]));
				writechar(',');
				writechar('x');
				writechar(')');
				break;
			case 10: // ; $12,x : $12+d+x
				writechar('$');
				writehex8(READBYTE(&program[1]));
				writechar(',');
				writechar('x');
				break;
			case 11: // ; $12,y : $12+d+y
				writechar('$');
				writehex8(READBYTE(&program[1]));
				writechar(',');
				writechar('y');
				break;
			case 12: // ; $1234,x : dbr+$1234+x
				writechar('$');
				writehex16(READWORD(&program[1]));
				writechar(',');
				writechar('x');
				break;
			case 13: // ; $1234,y : dbr+$1234+y
				writechar('$');
				writehex16(READWORD(&program[1]));
				writechar(',');
				writechar('y');
				break;
			case 14: // ; $123456,x : $123456+x
				writechar('$');
				writehex8(READBYTE(&program[3]));
				writehex16(READWORD(&program[1]));
				writechar(',');
				writechar('x');
				break;
			case 15: // ; +-$12 / $1234
				writechar('$');
				writehex16(READBYTE(&program[1]) + r.offset + 2);
				break;
			case 16: // ; +-$1234 / $1234
				writechar('$');
				writehex16(READWORD(&program[1]) + r.offset + 3);
				break;
			case 17: // ($1234)
				writechar('(');
				writechar('$');
				writehex16(READWORD(&program[1]));
				writechar(')');
				break;
			case 18: // ; ($12)
				writechar('(');
				writechar('$');
				writehex8(READBYTE(&program[1]));
				writechar(')');
				break;
			case 19: // ; [$12]
				writechar('[');
				writechar('$');
				writehex8(READBYTE(&program[1]));
				writechar(']');
				break;
			case 20: // ($1234,x)
				writechar('(');
				writechar('$');
				writehex16(READWORD(&program[1]));
				writechar(',');
				writechar('x');
				writechar(')');
				break;
			case 21: // ; s
				break;
			case 22: // ; d,s
				writechar('$');
				writehex8(READBYTE(&program[1]));
				writechar(',');
				writechar('s');
				break;
			case 23: // ; (d,s),y
				writechar('(');
				writechar('$');
				writehex8(READBYTE(&program[1]));
				writechar(')');
				writechar(',');
				writechar('y');
				break;
			case 24: // ; xyc - $1234
				writechar('$');
				writehex16(READWORD(&program[1]));
				break;
			case 25: // ; #$12 (Flag Operations)
				writechar('#');
				writechar('$');
				writehex8(READBYTE(&program[1]));
				break;
			case 26: // ; #$12,#$1234 (X-flag)
				writechar('#');
				writechar('$');
				if (r.status & 0x10) {
					// X-flag = 1
					writehex8(READBYTE(&program[1]));
					break;
				}
				writehex16(READWORD(&program[1]));
				break;
			case 27: // [$1234]
				writechar('[');
				writechar('$');
				writehex16(READWORD(&program[1]));
				writechar(']');
				break;
			}
			sz[x] = 0;
			while (x > 3 && sz[x-1] == ' ') x--;
			sz[x] = 0;

			strDelay += sz;

			if (r.ref) {
				CString strNew; strNew.Format(" -> L%d.1-%d"
					, 1U +r.refx.no
					, 0U +r.refx.n
					);
				while (strDelay.GetLength() < 20) strDelay += ' ';
				strDelay += strNew;
			}
			break;
		}
	}

	return strDelay;
}

CString CDebVw::GetHintProt0(UINT row, UINT col)
{
	const UINT index = row;

	CString strDelay;

	if (!(index < lineVec.GetCount())) return "?";
	Line &r = lineVec[index];

	switch (col) {
	case LCC_PBPC:
		if (r.label) break;

		strDelay.Format("%02X/%04X"
			, 0U +((r.offset & 0xFF0000) >> 16)
			, 0U +((r.offset & 0x00FFFF))
			);
		break;
	case LCC_COUNT:
		if (r.label) break;

		strDelay.Format("%ld", 0L +r.occur);
		break;
	case LCC_BC:
		{
			if (r.label) break;

			char sz[100] = {""};
			BYTE *program = (BYTE *)&r.execo;
			BYTE v = *program;
			BYTE n = SizeOfOp(program[0], r.status);
			for (BYTE i = 0; i < n; i++) {
				sprintf(strchr(sz, 0), "%02X ", 0U +program[i]);
			}
			strDelay = sz;
			break;
		}
	case LCC_INST:
		{
			if (r.label) {
				strDelay.Format("L%d.%d:"
					, 1U +r.labelx.no
					, 1U +r.labelx.branch
					);
				break;
			}

			BYTE v = (BYTE)r.execo;
			strDelay = CString((LPCSTR)&ocname[4*v], 4);
			BYTE *program = (BYTE *)&r.execo;

			char sz[100] = {""};
			char x = 0;
			switch (addrmode[v]) { // technique and knowledge from debug.asm
			case 0:
				break;
			case 1: // ; #$12,#$1234 (M-flag)
				writechar('#');
				writechar('$');
				if (r.status & 0x20) {
					// M-flag = 1 ---> 8-bit access
					writehex8(READBYTE(&program[1]));
					break;
				}
				// M-flag = 0 ---> 16-bit access
				writehex16(READWORD(&program[1]));
				break;
			case 2: // ; $1234 : db+$1234
				writechar('$');
				writehex16(READWORD(&program[1]));
				break;
			case 3: // ; $123456
				writechar('$');
				writehex8(READBYTE(&program[3]));
				writehex16(READWORD(&program[1]));
				break;
			case 4: // ; $12 : $12+d
				writechar('$');
				writehex8(READBYTE(&program[1]));
				break;
			case 5: // ; A
				writechar('A');
				break;
			case 6: // ; i
				break;
			case 7: // ; ($12),y : ($12+$13+d)+y
				writechar('(');
				writechar('$');
				writehex8(READBYTE(&program[1]));
				writechar(')');
				writechar(',');
				writechar('y');
				break;
			case 8: // ; [$12],y : [$12+$13+$14+d]+y
				writechar('[');
				writechar('$');
				writehex8(READBYTE(&program[1]));
				writechar(']');
				writechar(',');
				writechar('y');
				break;
			case 9: // ; ($12,x) : ($12+$13+d+x)
				writechar('(');
				writechar('$');
				writehex8(READBYTE(&program[1]));
				writechar(',');
				writechar('x');
				writechar(')');
				break;
			case 10: // ; $12,x : $12+d+x
				writechar('$');
				writehex8(READBYTE(&program[1]));
				writechar(',');
				writechar('x');
				break;
			case 11: // ; $12,y : $12+d+y
				writechar('$');
				writehex8(READBYTE(&program[1]));
				writechar(',');
				writechar('y');
				break;
			case 12: // ; $1234,x : dbr+$1234+x
				writechar('$');
				writehex16(READWORD(&program[1]));
				writechar(',');
				writechar('x');
				break;
			case 13: // ; $1234,y : dbr+$1234+y
				writechar('$');
				writehex16(READWORD(&program[1]));
				writechar(',');
				writechar('y');
				break;
			case 14: // ; $123456,x : $123456+x
				writechar('$');
				writehex8(READBYTE(&program[3]));
				writehex16(READWORD(&program[1]));
				writechar(',');
				writechar('x');
				break;
			case 15: // ; +-$12 / $1234
				writechar('$');
				writehex16(READBYTE(&program[1]) + r.offset + 2);
				break;
			case 16: // ; +-$1234 / $1234
				writechar('$');
				writehex16(READWORD(&program[1]) + r.offset + 3);
				break;
			case 17: // ($1234)
				writechar('(');
				writechar('$');
				writehex16(READWORD(&program[1]));
				writechar(')');
				break;
			case 18: // ; ($12)
				writechar('(');
				writechar('$');
				writehex8(READBYTE(&program[1]));
				writechar(')');
				break;
			case 19: // ; [$12]
				writechar('[');
				writechar('$');
				writehex8(READBYTE(&program[1]));
				writechar(']');
				break;
			case 20: // ($1234,x)
				writechar('(');
				writechar('$');
				writehex16(READWORD(&program[1]));
				writechar(',');
				writechar('x');
				writechar(')');
				break;
			case 21: // ; s
				break;
			case 22: // ; d,s
				writechar('$');
				writehex8(READBYTE(&program[1]));
				writechar(',');
				writechar('s');
				break;
			case 23: // ; (d,s),y
				writechar('(');
				writechar('$');
				writehex8(READBYTE(&program[1]));
				writechar(')');
				writechar(',');
				writechar('y');
				break;
			case 24: // ; xyc - $1234
				writechar('$');
				writehex16(READWORD(&program[1]));
				break;
			case 25: // ; #$12 (Flag Operations)
				writechar('#');
				writechar('$');
				writehex8(READBYTE(&program[1]));
				break;
			case 26: // ; #$12,#$1234 (X-flag)
				writechar('#');
				writechar('$');
				if (r.status & 0x10) {
					// X-flag = 1
					writehex8(READBYTE(&program[1]));
					break;
				}
				writehex16(READWORD(&program[1]));
				break;
			case 27: // [$1234]
				writechar('[');
				writechar('$');
				writehex16(READWORD(&program[1]));
				writechar(']');
				break;
			}
			sz[x] = 0;
			while (x > 3 && sz[x-1] == ' ') x--;
			sz[x] = 0;

			strDelay += sz;

			if (r.ref) {
				CString strNew; strNew.Format(" -> L%d.1-%d"
					, 1U +r.refx.no
					, 0U +r.refx.n
					);
				while (strDelay.GetLength() < 20) strDelay += ' ';
				strDelay += strNew;
			}
			break;
		}
	}

	return strDelay;
}

CString CDebVw::GetHintProt1(UINT row, UINT col)
{
	const UINT index = row;

	CString strDelay;

	if (!(index < lineVec.GetCount())) return "?";
	Line &r = lineVec[index];

	switch (col) {
	case LCC_PBPC:
		if (r.label) break;

		strDelay.Format("%04X"
			, 0U +((r.offset & 0x00FFFF))
			);
		break;
	case LCC_COUNT:
		if (r.label) break;

		strDelay.Format("%ld", 0L +r.occur);
		break;
	case LCC_BC:
		{
			if (r.label) break;

			char sz[100] = {""};
			BYTE *program = (BYTE *)&r.execo;
			BYTE v = *program;
			BYTE n = instapu[v].length;
			for (BYTE i = 0; i < n; i++) {
				sprintf(strchr(sz, 0), "%02X ", 0U +program[i]);
			}
			strDelay = sz;
			break;
		}
	case LCC_INST:
		{
			if (r.label) {
				strDelay.Format("L%d.%d:"
					, 1U +r.labelx.no
					, 1U +r.labelx.branch
					);
				break;
			}

			BYTE v = (BYTE)r.execo;
			strDelay = CString((LPCSTR)&spcnametab[8*v], 6);
			BYTE *program = (BYTE *)&r.execo;

			char sz[100] = {""};
			char x = 0;
			LPCSTR pszMap = (LPCSTR)&AddressTable[16*ArgumentTable[v]];
			for (char i = 0; i < 16; ) {
				char c = pszMap[i];
				if (c != '%') {
					sz[x] = c;
					x++;
					i++;
					continue;
				}
				i++;
				c = pszMap[i];
				i++;

				switch (c) {
					case '1':
						writehex8(READBYTE(&program[1]));
						break;
					case '2':
						writehex8(READBYTE(&program[2]));
						break;
					case '3':
						writehex4(READBYTE(&program[0]) >> 4);
						break;
					case '4':
						writehex4(READBYTE(&program[0]) >> 5);
						break;
					case '5':
						writehex16((r.offset + 2) & 0xFFFF);
						break;
					case '6':
						writechar('$');
						writechar((r.status & 20) ? '1' : '0');
						break;
					case '8':
						writehex16(READWORD(&program[1]) >> 3);
						break;
					case '9':
						writehex4(program[1] & 7);
						break;
					case 'A':
						writehex16(r.offset + 2 + (char)READBYTE(&program[1]));
						break;
					case 'B':
						writehex16(r.offset + 2 + (char)READBYTE(&program[2]));
						break;
				}
			}
			sz[x] = 0;
			while (x > 5 && sz[x-1] == ' ') x--;
			sz[x] = 0;

			strDelay += sz;

			if (r.ref) {
				CString strNew; strNew.Format(" -> L%d.1-%d"
					, 1U +r.refx.no
					, 0U +r.refx.n
					);
				while (strDelay.GetLength() < 20) strDelay += ' ';
				strDelay += strNew;
			}
			break;
		}
	}

	return strDelay;
}

CString CDebVw::GetHintAt(UINT row, UINT col)
{
	if (prot == 1)
		return GetHintProt1(row, col);
	if (prot == 2)
		return GetHintProt2(row, col);

	return GetHintProt0(row, col);
}

void CDebVw::OnLvnOdfinditem(NMHDR *pNMHDR, LRESULT *pResult)
{
	LPNMLVFINDITEM pFindInfo = reinterpret_cast<LPNMLVFINDITEM>(pNMHDR);

	*pResult = -1;
}

void CDebVw::OnEditCopy()
{
	CWaitCursor wc;

	CString str;
	{
		for (UINT i = 0; i < lineVec.GetCount(); i++) {
			CString strColumns[] = {
				GetHintAt(i, 0),
				GetHintAt(i, 1),
				GetHintAt(i, 2),
				GetHintAt(i, 3),
			};
			CString strLine;
			strLine.Format("|%7s|%10s|%-15s|%s\r\n"
				, (LPCTSTR)strColumns[0]
				, (LPCTSTR)strColumns[1]
				, (LPCTSTR)strColumns[2]
				, (LPCTSTR)strColumns[3]
				);
			str += strLine;
		}
	}

	if (OpenClipboard()) {
		size_t n = str.GetLength();
		EmptyClipboard();
		HGLOBAL hMem = GlobalAlloc(GHND, n + 1);
		LPVOID pvMem = GlobalLock(hMem);
		if (pvMem != NULL) {
			memcpy(pvMem, str, n + 1);
			GlobalUnlock(hMem);
			SetClipboardData(CF_TEXT, hMem);
		} else {
			GlobalFree(hMem);
		}
		CloseClipboard();
	}
}

void CDebVw::OnFileGraf() { }

void CDebVw::OnButtonSa1(UINT nID)
{
	switch (nID) {
		case ID_BUTTON_CPU:
			NewProt(0);
			break;
		case ID_BUTTON_APU:
			NewProt(1);
			break;
		case ID_BUTTON_SA1:
			NewProt(2);
			break;
	}
}

void CDebVw::OnUpdateButtonSa1(CCmdUI *pCmdUI)
{
	bool checked = false;
	switch (pCmdUI->m_nID) {
		case ID_BUTTON_CPU:
			checked = (prot == 0);
			break;
		case ID_BUTTON_APU:
			checked = (prot == 1);
			break;
		case ID_BUTTON_SA1:
			checked = (prot == 2);
			break;
	}
	pCmdUI->SetRadio(checked);
}

void CDebVw::NewProt(char protNew)
{
	prot = protNew;

//	ResetContents();
	OnFileConvert();
}
