
#include "StdAfx.h"
#include "SNESop.h"

namespace SNES {

PInstructionKind GetPikOf(BYTE v)
{
	switch (v) {
	case 0x90: // BCC
	case 0xB0: // BCS 
	case 0xF0: // BEQ 
	case 0x30: // BMI 
	case 0xD0: // BNE 
	case 0x10: // BPL 
	case 0x80: // BRA 
	case 0x82: // BRL 
	case 0x70: // BVS 
	case 0x50: // BVC
	case 0x4C: // JMP
	case 0x5C: // JMP
	case 0x6C: // JMP
	case 0x7C: // JMP
		return pikJxx;
	case 0x22: // JSL 
	case 0x20: // JSR
	case 0xFC: // JSR
		return pikSub;
	case 0x40: // RTI 
	case 0x6B: // RTL
	case 0x60: // RTS
		return pikRet;
	default:
		return pikNo;
	}
}

PISubKind GetPiskOf(BYTE v)
{
	switch (v) {
	case 0x22:
		return piskSubLong;
	case 0x20:
	case 0xFC:
		return piskSubShort;
	case 0x6B:
		return piskRetLong;
	case 0x60:
		return piskRetShort;
	case 0x40:
		return piskRetInterrupt;
	default:
		return piskNo;
	}
}

Instruction instruc[] = {
{"BRK", 5,2,8,0},
{"ORA",10,2,6,0},
{"COP", 5,2,8,0},
{"ORA",19,2,4,0},
{"TSB", 6,2,5,0},
{"ORA", 6,2,3,0},
{"ASL", 6,2,5,0},
{"ORA",12,2,6,0},
{"PHP", 5,1,3,0},
{"ORA", 0,2,2,0},
{"ASL", 1,1,2,0},
{"PHD", 5,1,4,0},
{"TSB",14,3,6,0},
{"ORA",14,3,4,0},
{"ASL",14,3,6,0},
{"ORA",17,4,5,0},
{"BPL", 2,2,2,1},
{"ORA",11,2,5,0},
{"ORA", 9,2,5,0},
{"ORA",20,2,7,0},
{"TRB", 6,2,5,0},
{"ORA", 7,2,4,0},
{"ASL", 7,2,6,0},
{"ORA",13,2,6,0},
{"CLC", 4,1,2,0},
{"ORA",16,3,4,0},
{"INC", 1,1,2,0},
{"TCS", 4,1,2,0},
{"TRB",14,3,6,0},
{"ORA",15,3,4,0},
{"ASL",15,3,7,0},
{"ORA",18,4,5,0},
{"JSR",14,3,6,1},
{"AND",10,2,6,0},
{"JSL",17,4,8,1},
{"AND",19,2,4,0},
{"BIT", 6,2,3,0},
{"AND", 6,2,3,0},
{"ROL", 6,2,5,0},
{"AND",12,2,6,0},
{"PLP", 5,1,4,0},
{"AND", 0,2,2,0},
{"ROL", 1,1,2,0},
{"PLD", 5,1,5,0},
{"BIT",14,3,4,0},
{"AND",14,3,4,0},
{"ROL",14,3,6,0},
{"AND",17,4,5,0},
{"BMI", 2,2,2,1},
{"AND",11,2,5,0},
{"AND", 9,2,5,0},
{"AND",20,2,7,0},
{"BIT", 7,2,4,0},
{"AND", 7,2,4,0},
{"ROL", 7,2,6,0},
{"AND",13,2,6,0},
{"SEC", 4,1,2,0},
{"AND",16,3,4,0},
{"DEC", 1,1,2,0},
{"TSC", 4,1,2,0},
{"BIT",15,3,4,0},
{"AND",15,3,4,0},
{"ROL",15,3,7,0},
{"AND",18,4,5,0},
{"RTI", 5,1,7,0}, // oĂTȂ̂Ŕr
{"EOR",10,2,6,0},
{"   ",-1,2,2,0},
{"EOR",19,2,4,0},
{"MVP",23,3,7,0},
{"EOR", 6,2,3,0},
{"LSR", 6,2,5,0},
{"EOR",12,2,6,0},
{"PHA", 5,1,3,0},
{"EOR", 0,2,2,0},
{"LSR", 1,1,2,0},
{"PHK", 5,1,3,0},
{"JMP",14,3,3,1},
{"EOR",14,3,4,0},
{"LSR",14,3,6,0},
{"EOR",17,4,5,0},
{"BVC", 2,2,2,1},
{"EOR",11,2,5,0},
{"EOR", 9,2,5,0},
{"EOR",20,2,7,0},
{"MVN",23,3,7,0},
{"EOR", 7,2,4,0},
{"LSR", 7,2,6,0},
{"EOR",13,2,6,0},
{"CLI", 4,1,2,0},
{"EOR",16,3,4,0},
{"PHY", 5,1,3,0},
{"TCD", 4,1,2,0},
{"JMP",17,4,4,1},
{"EOR",15,3,4,0},
{"LSR",15,3,7,0},
{"EOR",18,4,5,0},
{"RTS", 5,1,6,1},
{"ADC",10,2,6,0},
{"PER", 5,3,6,0},
{"ADC",19,2,4,0},
{"STZ", 6,2,3,0},
{"ADC", 6,2,3,0},
{"ROR", 6,2,5,0},
{"ADC",12,2,6,0},
{"PLA", 5,1,4,0},
{"ADC", 0,2,2,0},
{"ROR", 1,1,2,0},
{"RTL", 5,1,6,1},
{"JMP",21,3,5,1},
{"ADC",14,3,4,0},
{"ROR",14,3,6,0},
{"ADC",17,4,5,0},
{"BVS", 2,2,2,1},
{"ADC",11,2,5,0},
{"ADC", 9,2,5,0},
{"ADC",20,2,7,0},
{"STZ", 7,2,4,0},
{"ADC", 7,2,4,0},
{"ROR", 7,2,6,0},
{"ADC",13,2,6,0},
{"SEI", 4,1,2,0},
{"ADC",16,3,4,0},
{"PLY", 5,1,4,0},
{"TDC", 4,1,2,0},
{"JMP",22,3,6,1},
{"ADC",15,3,4,0},
{"ROR",15,3,7,0},
{"ADC",18,4,5,0},
{"BRA", 2,2,2,1},
{"STA",10,2,6,0},
{"BRL", 3,3,3,1},
{"STA",19,2,4,0},
{"STY", 6,2,3,0},
{"STA", 6,2,3,0},
{"STX", 6,2,3,0},
{"STA",12,2,6,0},
{"DEY", 4,1,2,0},
{"BIT", 0,2,2,0},
{"TXA", 4,1,2,0},
{"PHB", 5,1,3,0},
{"STY",14,3,4,0},
{"STA",14,3,4,0},
{"STX",14,3,4,0},
{"STA",17,4,5,0},
{"BCC", 2,2,2,1},
{"STA",11,2,6,0},
{"STA", 9,2,5,0},
{"STA",20,2,7,0},
{"STY", 7,2,4,0},
{"STA", 7,2,4,0},
{"STX", 8,2,4,0},
{"STA",13,2,6,0},
{"TYA", 4,1,2,0},
{"STA",16,3,5,0},
{"TXS", 4,1,2,0},
{"TXY", 4,1,2,0},
{"STZ",14,3,4,0},
{"STA",15,3,5,0},
{"STZ",15,3,5,0},
{"STA",18,4,5,0},
{"LDY", 0,2,2,0},
{"LDA",10,2,6,0},
{"LDX", 0,2,2,0},
{"LDA",19,2,4,0},
{"LDY", 6,2,3,0},
{"LDA", 6,2,3,0},
{"LDX", 6,2,3,0},
{"LDA",12,2,6,0},
{"TAY", 4,1,2,0},
{"LDA", 0,2,2,0},
{"TAX", 4,1,2,0},
{"PLB", 5,1,4,0},
{"LDY",14,3,4,0},
{"LDA",14,3,4,0},
{"LDX",14,3,4,0},
{"LDA",17,4,5,0},
{"BCS", 2,2,2,1},
{"LDA",11,2,5,0},
{"LDA", 9,2,5,0},
{"LDA",20,2,7,0},
{"LDY", 7,2,4,0},
{"LDA", 7,2,4,0},
{"LDX", 8,2,4,0},
{"LDA",13,2,6,0},
{"CLV", 4,1,2,0},
{"LDA",16,3,4,0},
{"TSX", 4,1,2,0},
{"TYX", 4,1,2,0},
{"LDY",15,3,4,0},
{"LDA",15,3,4,0},
{"LDX",16,3,4,0},
{"LDA",18,4,5,0},
{"CPY", 0,2,2,0},
{"CMP",10,2,6,0},
{"REP", 0,2,3,0},
{"CMP",19,2,4,0},
{"CPY", 6,2,3,0},
{"CMP", 6,2,3,0},
{"DEC", 6,2,5,0},
{"CMP",12,2,6,0},
{"INY", 4,1,2,0},
{"CMP", 0,2,2,0},
{"DEX", 4,1,2,0},
{"WAI", 4,1,3,0},
{"CPY",14,3,4,0},
{"CMP",14,3,4,0},
{"DEC",14,3,4,0},
{"CMP",17,4,5,0},
{"BNE", 2,2,2,1},
{"CMP",11,2,5,0},
{"CMP", 9,2,5,0},
{"CMP",20,2,7,0},
{"PEI", 5,2,6,0},
{"CMP", 7,2,4,0},
{"DEC", 7,2,6,0},
{"CMP",13,2,6,0},
{"CLD", 4,1,2,0},
{"CMP",16,3,4,0},
{"PHX", 5,1,3,0},
{"STP", 4,1,3,0},
{"JML",21,3,6,1},
{"CMP",15,3,4,0},
{"DEC",15,3,7,0},
{"CMP",18,4,5,0},
{"CPX", 0,2,2,0},
{"SBC",10,2,6,0},
{"SEP", 0,2,3,0},
{"SBC",19,2,4,0},
{"CPX", 6,2,3,0},
{"SBC", 6,2,3,0},
{"INC", 6,2,5,0},
{"SBC",12,2,6,0},
{"INX", 4,1,2,0},
{"SBC", 0,2,2,0},
{"NOP", 4,1,2,0},
{"XBA", 4,1,3,0},
{"CPX",14,3,4,0},
{"SBC",14,3,4,0},
{"INC",14,3,6,0},
{"SBC",17,4,5,0},
{"BEQ", 2,2,2,1},
{"SBC",11,2,5,0},
{"SBC", 9,2,5,0},
{"SBC",20,2,7,0},
{"PEA", 5,3,5,0},
{"SBC", 7,2,4,0},
{"INC", 7,2,6,0},
{"SBC",13,2,6,0},
{"SED", 4,1,2,0},
{"SBC",16,3,4,0},
{"PLX", 5,1,4,0},
{"XCE", 4,1,2,0},
{"JSR",22,3,6,1},
{"SBC",15,3,4,0},
{"INC",15,3,7,0},
{"SBC",18,4,5,0},
};

Instruction instapu[] = {

{NULL,0,1,0,0},
{NULL,0,1,0,1},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,0},
{NULL,0,1,0,1},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,0},
{NULL,0,3,0,0},
{NULL,0,3,0,1},
{NULL,0,1,0,0},
{NULL,0,1,0,1},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,1},
{NULL,0,2,0,0},
{NULL,0,1,0,1},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,1},
{NULL,0,1,0,0},
{NULL,0,1,0,1},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,1,0,1},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,0},
{NULL,0,3,0,0},
{NULL,0,3,0,1},
{NULL,0,1,0,0},
{NULL,0,1,0,1},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,1},
{NULL,0,2,0,0},
{NULL,0,1,0,1},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,0},
{NULL,0,1,0,1},
{NULL,0,1,0,0},
{NULL,0,1,0,1},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,1,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,1},
{NULL,0,1,0,1},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,1},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,0},
{NULL,0,1,0,1},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,1},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,1},
{NULL,0,1,0,1},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,1},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,1},
{NULL,0,1,0,1},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,2,0,0},
{NULL,0,3,0,0},
{NULL,0,2,0,0},
{NULL,0,1,0,0},
{NULL,0,1,0,0},
{NULL,0,2,0,0},
{NULL,0,1,0,0},

};

BYTE SizeOfOp(BYTE opec, BYTE ps)
{
	BYTE more = 0;
	switch (opec) {
		case 0x09: // ORA #
		case 0x29: // AND #
		case 0x49: // EOR #
		case 0x69: // ADC #
		case 0xA9: // LDA #
		case 0xC9: // CMP #
		case 0xE9: // SBC #
			if (ps & 0x20) {
				// M-flag = 1
				break;
			}
			// M-flag = 0
			more++;
			break;
		case 0xA0: // LDY #
		case 0xA2: // LDX #
		case 0xC0: // CPY #
		case 0xE0: // CPX #
			if (ps & 0x10) {
				// X-flag = 1
				break;
			}
			// X-flag = 0
			more++;
			break;
	}
	return instruc[opec].length + more;
}

BYTE SpecOfOp(BYTE opec)
{
	switch (opec) {
		case 0x90://BCC
		case 0xB0://BCS
		case 0xF0://BEQ
		case 0x30://BMI
		case 0xD0://BNE
		case 0x10://BPL
		case 0x80://BRA
		case 0x82://BRL
		case 0x50://BVC
		case 0x70://BVS
		case 0xDC://JML
		case 0x4C://JMP
		case 0x5C://JMP
		case 0x6C://JMP
		case 0x7C://JMP
			return 0x10;
		case 0x22://JSL
			return 0x01;
		case 0x20://JSR
		case 0xFC://JSR
			return 0x04;
		case 0x6B://RTL
			return 0x02;
		case 0x60://RTS
			return 0x08;
	}
	return 0x00;
}

BYTE SpecOfOpfaster(BYTE opec)
{
	switch (opec) {
		case 0x90://BCC
		case 0xB0://BCS
		case 0xF0://BEQ
		case 0x30://BMI
		case 0xD0://BNE
		case 0x10://BPL
	//	case 0x80://BRA
	//	case 0x82://BRL
		case 0x50://BVC
		case 0x70://BVS
		case 0xDC://JML
		case 0x4C://JMP
		case 0x5C://JMP
		case 0x6C://JMP
		case 0x7C://JMP
			return 0x10;
		case 0x22://JSL
			return 0x01;
		case 0x20://JSR
		case 0xFC://JSR
			return 0x04;
		case 0x6B://RTL
			return 0x02;
		case 0x60://RTS
			return 0x08;
	}
	return 0x00;
}

};

//  |	    |  |
// -+-------+--+-
// 0| #		|  |
// 1| A		|  |
// 2| r		|  |
// 3| rl	|  |
// 4| i		|  |
// 5| s		|  |
// 6| d		|  |
// 7| d,x	|  |
// 8| d,y	|  |
// 9|(d)	|  |
//10|(d,x)	|  |
//11|(d,y)	|  |
//12|[d]	|  |
//13|[d],y	|  |
//14| a		|  |
//15| a,x	|  |
//16| a,y	|  |
//17| al	|  |
//18| al,x	|  |
//19| d,s	|  |
//20|(d,s),y|  |
//21|(a)	|  |
//22|(a,x)	|  |
//23| xya	|  |
