
#include "StdAfx.h"
#include "ScriptServer.h"
#include "Server.h"
#include "fprintf.h"

// CScriptolet

void CScriptolet::Close()
{
	cUnkCommander.ReleaseDispatch();
	pUnkScript.Release();
	fReady = false;
}

bool CScriptolet::Load(CComBSTR bstrOLEScript, CComBSTR bstrText, CString strName)
{
	Close();

	strScriptoName = strName;
	HRESULT hr;
	if (FAILED(hr = pUnkScript.CoCreateInstance(bstrOLEScript, NULL, CLSCTX_INPROC_SERVER)))
		return false;
	if (FAILED(hr = pUnkScript->SetScriptSite(serverobj.GetMySite(false))))
		return false;
	CComQIPtr<IActiveScriptParse> pUnkScrParse = pUnkScript;
	if (pUnkScrParse == NULL)
		return false;
	if (FAILED(hr = pUnkScrParse->InitNew()))
		return false;
	if (FAILED(hr = pUnkScrParse->ParseScriptText(bstrText, NULL, NULL, NULL, 0, 0, 0, NULL, NULL)))
		return false;
	if (FAILED(hr = pUnkScript->AddNamedItem(L"external", 0 |SCRIPTITEM_NOCODE |SCRIPTITEM_ISVISIBLE)))
		return false;
	if (FAILED(hr = pUnkScript->SetScriptState(SCRIPTSTATE_STARTED)))
		return false;
	if (FAILED(hr = pUnkScript->SetScriptState(SCRIPTSTATE_CONNECTED)))
		return false;
	LPDISPATCH pUnkDisp = NULL;
	if (FAILED(hr = pUnkScript->GetScriptDispatch(NULL, &pUnkDisp)))
		return false;
	static LPWSTR pszNames[] = {
		L"OnCodeBreak",
		L"OnCodeBreakPostProcess",
	};
	for (char i = 0; i < 2; i++) {
		if (FAILED(hr = pUnkDisp->GetIDsOfNames(IID_NULL, &pszNames[i], 1, 0, &cDispID[i]))) {
			if (hr != DISP_E_UNKNOWNNAME)
				return false;
		}
	}
	cUnkCommander.AttachDispatch(pUnkDisp);

	fReady = true;
	return true;
}

bool CScriptolet::LoadFile(CComBSTR bstrOLEScript, LPCTSTR pszFileName)
{
	const int ch = 100*1024;
	static BYTE sz[ch + 1];
	ZeroMemory(sz, ch);

	FILE *f = _tfopen(pszFileName, "rt");
	if (f == NULL)
		return false;
	int r = fread(sz, 1, ch, f);
	fclose(f);
	CComBSTR bstrText(r, (LPCSTR)sz);
	return Load(bstrOLEScript, bstrText, pszFileName);
}

int CScriptolet::GetState()
{
	if (pUnkScript == NULL)
		return SCRLETS_UNLOADED;
	if (!fReady)
		return SCRLETS_BROKEN;
	SCRIPTSTATE ss;
	HRESULT hr;
	if (FAILED(hr = pUnkScript->GetScriptState(&ss)))
		return SCRLETS_INTERNAL_ERR;
	if (ss == SCRIPTSTATE_CONNECTED)
		return SCRLETS_OK;
	return SCRLETS_UNKNOWN;
}

// CScriptoServer

bool CScriptoServer::Sync()
{
	size_t i;
	for (i = 0; i < MAX_SCRIPTO; i++) {
		scrlet[i].Close();
	}
	for (i = 0; i < MAX_SCRIPTO && i < scriptoVec.size(); i++) {
		if (scriptoVec[i].fDisabled) continue;
		scrlet[i].LoadFile((LPCTSTR)scriptoVec[i].strEngine, scriptoVec[i].strFile);
	}

	return true;
}

void CScriptoServer::CodeBreakNow(DWORD addr)
{
	static const BYTE parms[] = VTS_I4;
	size_t i;
	for (i = 0; i < MAX_SCRIPTO; i++) {
		CScriptolet &r = scrlet[i];
		if (!r.fReady) continue;
		TRY {
			if (r.cUnkCommander.m_lpDispatch == NULL || r.cDispID[0] == DISPID_UNKNOWN) continue;
			r.cUnkCommander.InvokeHelper(r.cDispID[0], DISPATCH_METHOD, VT_EMPTY, NULL, parms, addr);
		} CATCH(COleDispatchException, e) {
			r.fReady = false;
		} END_CATCH;
	}
}

void CScriptoServer::CodeBreakPostProcNow(DWORD addrOld, DWORD addrNew)
{
	static const BYTE parms[] = VTS_I4 VTS_I4;
	size_t i;
	for (i = 0; i < MAX_SCRIPTO; i++) {
		CScriptolet &r = scrlet[i];
		if (!r.fReady) continue;
		TRY {
			if (r.cUnkCommander.m_lpDispatch == NULL || r.cDispID[1] == DISPID_UNKNOWN) continue;
			r.cUnkCommander.InvokeHelper(r.cDispID[1], DISPATCH_METHOD, VT_EMPTY, NULL, parms, addrOld, addrNew);
		} CATCH(COleDispatchException, e) {
			r.fReady = false;
		} END_CATCH;
	}
}

void CScriptoServer::Close()
{
	size_t i;
	for (i = 0; i < MAX_SCRIPTO; i++) {
		scrlet[i].Close();
	}
}
