
#include "stdafx.h"
#include "Server.h"
#include "IDEFrame.h"
#include "fprintf.h"
#include "fprintfi.h"
#include "ServerMsgob.h"

using namespace OSP;

extern CIDEFrame frameide;

extern "C" unsigned char *romdata;
extern "C" unsigned int infoloc;

namespace {

CString GetAppDir()
{
	return OSP_GetDir(OSP_GetModuleFileName(NULL));
}

CString RelativeJoinPath(CString strBase, CString strAdd)
{
	if (strAdd.Left(2).Compare(".\\") == 0)
		return OSP_JoinPath(strBase, strAdd);
	return strAdd;
}

};

// CServer

CServer serverobj;

CServer::CServer()
{
	EnableAutomation();
}

CServer::~CServer()
{

}

void CServer::OnFinalRelease()
{
	CCmdTarget::OnFinalRelease();
}

LPDISPATCH CServer::Message(LPCSTR pszMessage, LPCTSTR pszSrc, int severity)
{
	int refid;
	if (frameide.wndCustDock.frame.Avail()) {
		frameide.wndCustDock.frame.vw.Message(pszMessage, pszSrc, severity, refid);

		if (refid != -1) {
			CServerMsgob *ob = new CServerMsgob(refid);
			return ob->GetIDispatch(false);
		}
	}

	return NULL;
}

BEGIN_MESSAGE_MAP(CServer, CCmdTarget)
END_MESSAGE_MAP()

BEGIN_DISPATCH_MAP(CServer, CCmdTarget)
	DISP_FUNCTION(CServer, "Message", U_Message, VT_DISPATCH, VTS_BSTR VTS_BSTR VTS_I4)
	DISP_FUNCTION(CServer, "AddCodeBreak", U_AddCodeBreak, VT_DISPATCH, VTS_I4)
	DISP_FUNCTION(CServer, "RemoveCodeBreak", U_RemoveCodeBreak, VT_DISPATCH, VTS_I4)
	DISP_FUNCTION(CServer, "ReadMem8", U_ReadMem8, VT_I1, VTS_I4)
	DISP_FUNCTION(CServer, "ReadMem16", U_ReadMem16, VT_I2, VTS_I4)
	DISP_FUNCTION(CServer, "ReadMem32", U_ReadMem32, VT_I4, VTS_I4)
//	DISP_FUNCTION(CServer, "GetReg", U_GetReg, VT_I4, VTS_I4)
	DISP_FUNCTION(CServer, "CaptureMemoryToFile", U_CaptureMemoryToFile, VT_BOOL, VTS_I4 VTS_I4 VTS_BSTR)
	DISP_FUNCTION(CServer, "sprintf", U_sprintf, VT_BSTR, VTS_WBSTR VTS_VARIANT)
	DISP_FUNCTION(CServer, "EnableSingleStep", U_EnableSingleStep, VT_EMPTY, VTS_NONE)
	DISP_FUNCTION(CServer, "DisableSingleStep", U_DisableSingleStep, VT_EMPTY, VTS_NONE)
	DISP_FUNCTION(CServer, "EnablePostProcess", U_EnablePostProcess, VT_EMPTY, VTS_NONE)
	DISP_FUNCTION(CServer, "DisablePostProcess", U_DisablePostProcess, VT_EMPTY, VTS_NONE)
	DISP_FUNCTION(CServer, "CaptureMemoryToVar", U_CaptureMemoryToVar, VT_VARIANT, VTS_I4 VTS_I4)

	DISP_PROPERTY_EX(CServer, "xa", U_xa, U_seterr, VT_I2)
	DISP_PROPERTY_EX(CServer, "xdb", U_xdb, U_seterr, VT_I1)
	DISP_PROPERTY_EX(CServer, "xpb", U_xpb, U_seterr, VT_I1)
	DISP_PROPERTY_EX(CServer, "xs", U_xs, U_seterr, VT_I2)
	DISP_PROPERTY_EX(CServer, "xd", U_xd, U_seterr, VT_I2)
	DISP_PROPERTY_EX(CServer, "xx", U_xx, U_seterr, VT_I2)
	DISP_PROPERTY_EX(CServer, "xy", U_xy, U_seterr, VT_I2)
	DISP_PROPERTY_EX(CServer, "xp", U_xp, U_seterr, VT_I1)
	DISP_PROPERTY_EX(CServer, "CasetteName", U_CasetteName, U_seterr, VT_BSTR)
	DISP_PROPERTY_EX(CServer, "PrevAddr", U_PrevAddr, U_seterr, VT_I4)
END_DISPATCH_MAP()

// {354E73BE-391A-47A8-8D26-7893674B6FAB}
static const IID IID_IServer =
{ 0x354E73BE, 0x391A, 0x47A8, { 0x8D, 0x26, 0x78, 0x93, 0x67, 0x4B, 0x6F, 0xAB } };

BEGIN_INTERFACE_MAP(CServer, CCmdTarget)
	INTERFACE_PART(CServer, IID_IServer, Dispatch)
	INTERFACE_PART(CServer, IID_IActiveScriptSite, MySite)
	INTERFACE_PART(CServer, IID_IActiveScriptSiteWindow, MySiteWindow)
END_INTERFACE_MAP()

/*//////////////////////////////////////////////////////////////////////////
// Published features
//////////////////////////////////////////////////////////////////////////*/

LPDISPATCH CServer::U_Message(LPCSTR pszMessage, LPCTSTR pszSrc, int severity)
{
	return Message(pszMessage, CString("@") + pszSrc, severity);
}

LPDISPATCH CServer::U_AddCodeBreak(int addr)
{
	DWORD x;
	for (x = 0; x < MAX_CODEBREAK; x++) {
		if (breakCode[x].addr != addr) continue;

		breakCode[x].nRefs++;
		return NULL;
	}
	for (x = 0; x < MAX_CODEBREAK; x++) {
		if (breakCode[x].nRefs != 0) continue;

		breakCode[x].addr = addr;
		breakCode[x].nRefs++;
		return NULL;
	}

//	Message("No more avail code break entries.", "%fprintf", 3);
	AfxThrowOleDispatchException(1, "No more avail code break entries.");
	return NULL;
}

void CServer::U_RemoveCodeBreak(int addr)
{
	DWORD x;
	for (x = 0; x < MAX_CODEBREAK; x++) {
		if (breakCode[x].addr != addr) continue;

		breakCode[x].nRefs--;
		return;
	}

	return;
}

DWORD CServer::U_ReadMem8(int addr)
{
	fprintfi_addr = addr;
	fprintfi_read_mem8();
	return fprintfi_read & 0x000000FF;
}

DWORD CServer::U_ReadMem16(int addr)
{
	fprintfi_addr = addr;
	fprintfi_read_mem16();
	return fprintfi_read & 0x0000FFFF;
}

DWORD CServer::U_ReadMem32(int addr)
{
	fprintfi_addr = addr;
	fprintfi_read_mem32();
	return fprintfi_read;
}

extern "C" DWORD xa,xd,xdb,xe,xp,xpb,xpc,xs,xx,xy, xp;

int CServer::U_GetReg(int what)
{
	switch (what) {
	case  0: return (xa  & 0xFFFF);
	case  1: return (xdb & 0x00FF);
	case  2: return (xpb & 0x00FF);
	case  3: return (xs  & 0xFFFF);
	case  4: return (xd  & 0xFFFF);
	case  5: return (xx  & 0xFFFF);
	case  6: return (xy  & 0xFFFF);
	case  7: return (xp  & 0x00FF);
	}
	return 0;
}

void CServer::U_seterr(DWORD)
{
	AfxThrowNotSupportedException();
}

WORD CServer::U_xa()  { return static_cast<WORD>(xa  & 0xFFFF); }
BYTE CServer::U_xdb() { return static_cast<BYTE>(xdb & 0x00FF); }
BYTE CServer::U_xpb() { return static_cast<BYTE>(xpb & 0x00FF); }
WORD CServer::U_xs()  { return static_cast<WORD>(xs  & 0xFFFF); }
WORD CServer::U_xd()  { return static_cast<WORD>(xd  & 0xFFFF); }
WORD CServer::U_xx()  { return static_cast<WORD>(xx  & 0xFFFF); }
WORD CServer::U_xy()  { return static_cast<WORD>(xy  & 0xFFFF); }
BYTE CServer::U_xp()  { return static_cast<BYTE>(xp  & 0x00FF); }

BSTR CServer::U_CasetteName()
{
	CString strName((LPCSTR)fprintf_infoloc, 20);
	strName.Trim();
	return strName.AllocSysString();
}

BOOL CServer::U_CaptureMemoryToFile(DWORD addrFrom, DWORD addrTo, LPCTSTR pszNewFile)
{
	if (0xFFFFFF < addrFrom)
		AfxThrowOleDispatchException(1, "ERR! (0xFFFFFF < addrFrom)");
	if (0xFFFFFF < addrTo)
		AfxThrowOleDispatchException(1, "ERR! (0xFFFFFF < addrTo)");
	if (addrTo < addrFrom)
		AfxThrowOleDispatchException(1, "ERR! (addrTo < addrFrom)");

	FILE *f = _tfopen(RelativeJoinPath(GetAppDir(), pszNewFile), "wb");
	if (f == NULL)
		return false;

	static const int chBuffer = 4096;
	static BYTE szBuffer[chBuffer];

	int i = 0;
	int err = 0;

	while (addrFrom <= addrTo) {
		if (i == chBuffer) {
			if (fwrite(szBuffer, i, 1, f) != 1) {
				err++;
				break;
			}
			i = 0;
		}

		fprintfi_addr = addrFrom;
		fprintfi_read_mem8();
		szBuffer[i] = (BYTE)fprintfi_read;

		i++;
		addrFrom++;
	}

	if (err == 0 && i != 0) {
		if (fwrite(szBuffer, i, 1, f) != 1) {
			err++;
		}
	}

	fclose(f);
	return (err == 0);
}

struct AccessToXArray {
	// 
	virtual HRESULT GetElement(int index, VARIANT &var) = NULL;
};
struct AccessToVBArray : AccessToXArray {
	// 
	HRESULT hr;
	SAFEARRAY *psa;

	// 
	virtual HRESULT GetElement(int index, VARIANT &var)
	{
		LONG indices[2] = {index, 0};
		if (FAILED(hr = SafeArrayGetElement(psa, indices, &var)))
			return hr;
		return S_OK;
	}
};
struct AccessToJSArray : AccessToXArray {
	// 
	HRESULT hr;
	LPDISPATCH pdispVal;
	CArray<CComVariant> varVec;

	// 
	HRESULT Explode()
	{
		UINT iArgErr = 0;
		DISPPARAMS parms;
		ZeroMemory(&parms, sizeof(parms));
		CComVariant varNewEnum;
		if (FAILED(hr = pdispVal->Invoke(DISPID_NEWENUM, IID_NULL, 0, DISPATCH_METHOD, &parms, &varNewEnum, NULL, &iArgErr))) {
			return hr;
		}
		CComQIPtr<IEnumVARIANT> pNewEnum(varNewEnum.punkVal);
		if (pNewEnum == NULL)
			return E_NOTIMPL;
		int size = 20;
		int base = 0;
		while (true) {
			varVec.SetSize(size);
			ULONG celt = 0;
			if (FAILED(hr = pNewEnum->Next(20, &varVec.ElementAt(base), &celt)))
				return hr;
			if (hr == S_FALSE || celt != size - base) {
				varVec.SetSize(base + celt);
				break;
			}
			base = size;
			size+= size;
		}
		return S_OK;
	}

	// 
	virtual HRESULT GetElement(int index, VARIANT &var)
	{
		if (index < varVec.GetSize()) {
			VariantCopy(&var, &varVec[index]);
			return S_OK;
		}
		return OLE_E_ENUM_NOMORE;
	}
};

struct LikeVararg {
	// 
	BYTE buffer[1024];
	size_t pos;

	// 
	LikeVararg()
	{
		pos = 0;
	}
	// 
	bool Write(const void *pv, size_t size)
	{
		if (!(pos < sizeof(buffer) + size))
			return false;
		memcpy(buffer + pos, pv, size);
		pos += size;
		return true;
	}
};

BSTR CServer::U_sprintf(LPCOLESTR format, VARIANT &params)
{
#if defined(_M_IX86) && defined(_WIN32)
	HRESULT hr;
	VARIANT *pp = &params;
	SAFEARRAY *psa = NULL;
	LPDISPATCH pdispVal = NULL;
	AccessToVBArray vbsa;
	AccessToJSArray jsa;
	if (pp->vt == (0 |VT_BYREF |VT_VARIANT)) {
		pp = pp->pvarVal;
	}
	if (pp->vt == (0 |VT_ARRAY |VT_VARIANT)) {
		vbsa.psa = psa = pp->parray;
	}
	if (pp->vt == (0 |VT_DISPATCH)) {
		jsa.pdispVal = pdispVal = pp->pdispVal;
	}
	if (psa == NULL && pdispVal == NULL)
		AfxThrowOleDispatchException(1, _T("Second parameter has to be an array"));
	if (psa != NULL) {
		VARTYPE vt;
		if (FAILED(hr = SafeArrayGetVartype(psa, &vt)) || vt != VT_VARIANT)
			AfxThrowOleDispatchException(1, _T("Second parameter has to be variant array"));
	} else {
		if (FAILED(hr = jsa.Explode()))
			AfxThrowOleDispatchException(1, _T("Second parameter has to be DISPID_NEWENUM capable object"));
	}
	AccessToXArray &xa = (psa != NULL) ? (AccessToXArray &)vbsa : (AccessToXArray &)jsa;

	if (format == NULL || *format == NULL)
		return NULL;

	int indexParm = 0;

	CArray<CComBSTR> textVec;

	LikeVararg likeva;

	LPCOLESTR form = format;
	wchar_t c, cc, ccc;
	while (*form != 0) {
		c = *form;
		form++;
		if (c == L'%') {
			cc = *form;
			if (cc == 0)
				break;
			if (cc == '%') {
				form++;
			} else {
				// %[flags] [width] [.precision] [{h |l |I | I32 | I64}]type
				ccc = *form;
				if (ccc == 0)
					break;
				if (false
					|| ccc == '-'
					|| ccc == '+'
					|| ccc == '0'
					|| ccc == '#'
				) {
					form++;
				}

				while (true) {
					ccc = *form;
					if ('0' <= ccc && ccc <= '9') {
						form++;
						continue;
					}
					break;
				}

				ccc = *form;
				if (ccc == 0)
					break;
				if (ccc == '.') {
					form++;
					ccc = *form;
					if (ccc == 0)
						break;
					while (true) {
						ccc = *form;
						if ('0' <= ccc && ccc <= '9') {
							form++;
							continue;
						}
						break;
					}
				}

				wchar_t sp = *form;
				if (sp == 0)
					break;
				if (false);
				else if (sp == 'h') {
					sp = 2;
					form++;
				}
				else if (sp == 'l') {
					sp = 4;
					form++;
				}
				else if (sp == 'I') {
					form++;
					if (false);
					else if (form[0] == '3' && form[1] == '2') {
						sp = 4;
						form += 2;
					}
					else if (form[0] == '6' && form[1] == '4') {
						sp = 8;
						form += 2;
					}
				}
				else {
					sp = 0;
				}

				if (sp != 0) {
					AfxThrowOleDispatchException(1, _T("Argument size specifier not supported"));
				}

				CComVariant var;
				if (FAILED(hr = xa.GetElement(indexParm, var)))
					AfxThrowOleDispatchException(1, _T("Too few parameters"));
				indexParm++;

				ccc = *form;
				if (ccc == 0)
					break;
				form++;
				char pp = 0;
				switch (ccc) {
					case 'c':
						{
							hr = var.ChangeType(VT_I2);
							pp = 'c';

							break;
						}
					case 'd':
					case 'i':
					case 'o':
					case 'u':
					case 'x':
					case 'X':
						{
							hr = var.ChangeType(VT_I4);
							pp = 'd';
							break;
						}
					case 'e':
					case 'E':
					case 'f':
					case 'g':
					case 'G':
						{
							hr = var.ChangeType(VT_R8);
							pp = 'e';
							break;
						}
					case 's':
						{
							hr = var.ChangeType(VT_BSTR);
							pp = 's';
							break;
						}
					default:
						{
							AfxThrowOleDispatchException(1, _T("Contains unknown external type field"));
						}
				}
				if (FAILED(hr))
					AfxThrowOleDispatchException(1, _T("Contains invalid parameter for conversion"));
				switch (pp) {
					case 'c': // WIDE
						{
							DWORD x = var.iVal;
							likeva.Write(&x, 4);
							break;
						}
					case 'C': // ANSI
						{
							likeva.Write(&var.cVal, 1);
							break;
						}
					case 'd':
						{
							likeva.Write(&var.intVal, 4);
							break;
						}
					case 'e':
						{
							likeva.Write(&var.dblVal, 8);
							break;
						}
					case 's':
						{
							textVec.Add(var.bstrVal);
							LPCWSTR psz = textVec[textVec.GetSize() -1];
							likeva.Write(&psz, sizeof(LPCWSTR));
							break;
						}
					default:
						{
							AfxThrowOleDispatchException(1, _T("Contains unknown internal type field"));
						}
				}
			}
		}
	}

	va_list marker = (char *)likeva.buffer;
	CStringW strNew;
	strNew.FormatV(format, marker);
	va_end(marker);

	return strNew.AllocSysString();
#else
#error HAS TO BE _M_IX86
#endif
}

void CServer::U_EnableSingleStep()
{
	if (refSingleStep != 0x7FFFFFFF)
		refSingleStep++;
}
void CServer::U_DisableSingleStep()
{
	if (refSingleStep != 0)
		refSingleStep--;
}
void CServer::U_EnablePostProcess()
{
	if (refPostProcess != 0x7FFFFFFF)
		refPostProcess++;
}
void CServer::U_DisablePostProcess()
{
	if (refPostProcess != 0)
		refPostProcess--;
}

DWORD CServer::U_PrevAddr()
{
	return fprintfc_old_offset;
}

VARIANT CServer::U_CaptureMemoryToVar(DWORD addrFrom, DWORD addrTo)
{
	if (0xFFFFFF < addrFrom)
		AfxThrowOleDispatchException(1, "ERR! (0xFFFFFF < addrFrom)");
	if (0xFFFFFF < addrTo)
		AfxThrowOleDispatchException(1, "ERR! (0xFFFFFF < addrTo)");
	if (addrTo < addrFrom)
		AfxThrowOleDispatchException(1, "ERR! (addrTo < addrFrom)");

	HRESULT hr;
	COleVariant vt;
	SAFEARRAYBOUND sab;
	sab.lLbound = 0;
	sab.cElements = addrTo - addrFrom + 1;
	LPSAFEARRAY psa = SafeArrayCreate(VT_UI1, 1, &sab);
	if (psa != NULL) {
		void *pv;
		if (SUCCEEDED(hr = SafeArrayAccessData(psa, &pv))) {
			BYTE *szBuffer = (BYTE *)pv;
			int i = 0;
			while (addrFrom <= addrTo) {
				fprintfi_addr = addrFrom;
				fprintfi_read_mem8();
				*szBuffer = (BYTE)fprintfi_read;

				szBuffer++;
				addrFrom++;
			}
			hr = SafeArrayUnaccessData(psa);

			vt.vt = 0 |VT_UI1 |VT_ARRAY;
			vt.parray = psa;
			return vt.Detach();
		}
		SafeArrayDestroy(psa);
	}
	AfxThrowMemoryException();
	return vt.Detach();
}

/*//////////////////////////////////////////////////////////////////////////
// IActiveScriptSite
//////////////////////////////////////////////////////////////////////////*/

ULONG FAR EXPORT CServer::XMySite::AddRef()
{
	METHOD_PROLOGUE(CServer, MySite)
	return pThis->ExternalAddRef();
}

ULONG FAR EXPORT CServer::XMySite::Release()
{
	METHOD_PROLOGUE(CServer, MySite)
	return pThis->ExternalRelease();
}

HRESULT FAR EXPORT CServer::XMySite::QueryInterface(
	REFIID iid, void FAR* FAR* ppvObj)
{
	METHOD_PROLOGUE(CServer, MySite)
	return (HRESULT)pThis->ExternalQueryInterface(&iid, ppvObj);
}

HRESULT CServer::XMySite::GetLCID(LCID *)
{
	return E_NOTIMPL;
}

HRESULT CServer::XMySite::GetItemInfo(LPCOLESTR pstrName, DWORD nReturnMask, IUnknown **ppiunkItem, ITypeInfo **ppti)
{
	METHOD_PROLOGUE(CServer, MySite)

	if (ppiunkItem != NULL) *ppiunkItem = NULL;
	if (ppti != NULL) *ppti = NULL;

	if (wcsicmp(pstrName, L"external") == 0) {
		if (ppiunkItem != NULL) {
			*ppiunkItem = this;
			AddRef();
		}
		return S_OK;
	}
	return TYPE_E_ELEMENTNOTFOUND;
}

HRESULT CServer::XMySite::GetDocVersionString(BSTR *pbstrVersion)
{
	return E_NOTIMPL;
}

HRESULT CServer::XMySite::OnScriptTerminate(const VARIANT *pvarResult, const EXCEPINFO *pexcepinfo)
{
	return S_OK;
}

HRESULT CServer::XMySite::OnStateChange(SCRIPTSTATE ssScriptState)
{
	return S_OK;
}

HRESULT CServer::XMySite::OnScriptError(IActiveScriptError *pscripterror)
{
	METHOD_PROLOGUE(CServer, MySite)

	HRESULT hr;
	EXCEPINFO ei;
	ZeroMemory(&ei, sizeof(ei));
	DWORD nSourceContext;
	ULONG nLineNumber;
	LONG nCharacterPosition;
	if (SUCCEEDED(hr = pscripterror->GetExceptionInfo(&ei)) && SUCCEEDED(pscripterror->GetSourcePosition(&nSourceContext, &nLineNumber, &nCharacterPosition))) {
		CString strErr;
		strErr.Format(
			"[%s] %s\n"
			"----------\n"
			"Code       : %d (0x%04lX), %d (0x%08lX)\n"
			"Description: \"%s\"\n"
			"Source     : \"%s\"\n"
			"Location   : Line=%u, Position=%u\n"
			, (LPCTSTR)CString(ei.bstrSource)
			, (LPCTSTR)CString(ei.bstrDescription)
			, (ULONG)ei.wCode, (ULONG)ei.wCode, (ULONG)ei.scode, (ULONG)ei.scode
			, (LPCSTR)CString(ei.bstrDescription)
			, (LPCSTR)CString(ei.bstrSource)
			, (UINT)1+nLineNumber, (UINT)1+nCharacterPosition
			);

		serverobj.Message(strErr, "%fprintf", 3);

		SysFreeString(ei.bstrSource);
		SysFreeString(ei.bstrDescription);
		SysFreeString(ei.bstrHelpFile);

		TRACE1("%s", strErr);
		return E_ABORT;
	}
	return S_OK;
}

HRESULT CServer::XMySite::OnEnterScript()
{
	return S_OK;
}

HRESULT CServer::XMySite::OnLeaveScript()
{
	return S_OK;
}

IActiveScriptSite *CServer::GetMySite(bool fAddRef)
{
	if (fAddRef)
		m_xMySite.AddRef();

	return &m_xMySite;
}

/*//////////////////////////////////////////////////////////////////////////
// IActiveScriptSiteWindow
//////////////////////////////////////////////////////////////////////////*/

ULONG FAR EXPORT CServer::XMySiteWindow::AddRef()
{
	METHOD_PROLOGUE(CServer, MySiteWindow)
	return pThis->ExternalAddRef();
}

ULONG FAR EXPORT CServer::XMySiteWindow::Release()
{
	METHOD_PROLOGUE(CServer, MySiteWindow)
	return pThis->ExternalRelease();
}

HRESULT FAR EXPORT CServer::XMySiteWindow::QueryInterface(
	REFIID iid, void FAR* FAR* ppvObj)
{
	METHOD_PROLOGUE(CServer, MySiteWindow)
	return (HRESULT)pThis->ExternalQueryInterface(&iid, ppvObj);
}

HRESULT CServer::XMySiteWindow::GetWindow(HWND *phwnd)
{
	*phwnd = AfxGetMainWnd()->GetSafeHwnd();
	return S_OK;
}

HRESULT CServer::XMySiteWindow::EnableModeless(BOOL fEnable)
{
	EnableWindow(AfxGetMainWnd()->GetSafeHwnd(), fEnable);
	return S_OK;
}
