
#include "StdAfx.h"
#include "Util.h"

extern "C" BYTE cgram[];

namespace SNESUtil
{
	COLORREF getPaletteAt(BYTE index)
	{
		static const BYTE table[32] = {
			0,8,16,24,32,41,49,57,65,74,82,90,98,106,115,123,131,139,148,156,164,172,180,189,197,205,213,222,230,238,246,255,
		};
		WORD v = ((WORD *)cgram)[index];
		return RGB(
			table[(v >> 0) & 31],
			table[(v >> 5) & 31],
			table[(v >>10) & 31]
			);
	}

	void getPaletteAt(BYTE index, RGBQUAD &rgb)
	{
		static const BYTE table[32] = {
			0,8,16,24,32,41,49,57,65,74,82,90,98,106,115,123,131,139,148,156,164,172,180,189,197,205,213,222,230,238,246,255,
		};
		WORD v = ((WORD *)cgram)[index];
		rgb.rgbRed   = table[(v >> 0) & 31];
		rgb.rgbGreen = table[(v >> 5) & 31];
		rgb.rgbBlue  = table[(v >>10) & 31];
		rgb.rgbReserved = 0;
	}
};

namespace SNESUtil
{
	void convert8x8_2_to_8(BYTE src[16], BYTE cvt[64])
	{
		static const BYTE table[] = {
			// 7,   6,   5,   4,   3,   2,   1,   0,
			0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01,
		};
		BYTE *pix = cvt;
		for (BYTE y=0; y<8; y++) {
			for (BYTE x=0; x<8; x++) {
				*pix = 0
					| ((src[   +2*y   ] & table[x]) ? 1 : 0)
					| ((src[   +2*y +1] & table[x]) ? 2 : 0)
					;

				pix++;
			}
		}
	}

	void convert8x8_4_to_8(BYTE src[32], BYTE cvt[64])
	{
		static const BYTE table[] = {
			// 7,   6,   5,   4,   3,   2,   1,   0,
			0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01,
		};
		BYTE *pix = cvt;
		for (BYTE y=0; y<8; y++) {
			for (BYTE x=0; x<8; x++) {
				*pix = 0
					| ((src[   +2*y   ] & table[x]) ? 1 : 0)
					| ((src[   +2*y +1] & table[x]) ? 2 : 0)
					| ((src[16 +2*y   ] & table[x]) ? 4 : 0)
					| ((src[16 +2*y +1] & table[x]) ? 8 : 0)
					;

				pix++;
			}
		}
	}

	void convert16x8_4_to_8(BYTE src[64], BYTE cvt[128])
	{
		static const BYTE tablex[] = {
			// 7,   6,   5,   4,   3,   2,   1,   0,
			0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01,
		};
		static const BYTE tabletx[] = {0,1};
		BYTE *pix = cvt;
		for (BYTE y=0; y<8; y++) {
			for (BYTE tx=0; tx<2; tx++) {
				for (BYTE x=0; x<8; x++) {
					*pix = 0
						| ((src[   +4*y    +tabletx[tx]] & tablex[x]) ? 1 : 0)
						| ((src[   +4*y +2 +tabletx[tx]] & tablex[x]) ? 2 : 0)
						| ((src[32 +4*y    +tabletx[tx]] & tablex[x]) ? 4 : 0)
						| ((src[32 +4*y +2 +tabletx[tx]] & tablex[x]) ? 8 : 0)
						;

					pix++;
				}
			}
		}
	}

	void convert16x16_4_to_8(BYTE src[128], BYTE cvt[256])
	{
		static const BYTE tablex[] = {
			// 7,   6,   5,   4,   3,   2,   1,   0,
			0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01,
		};
		static const BYTE tabletx[] = {1,0};
		BYTE *pix = cvt;
		for (BYTE y=0; y<16; y++) {
			for (BYTE tx=0; tx<2; tx++) {
				for (BYTE x=0; x<8; x++) {
					*pix = 0
						| ((src[   +4*y    +tabletx[tx]] & tablex[x]) ? 1 : 0)
						| ((src[   +4*y +2 +tabletx[tx]] & tablex[x]) ? 2 : 0)
						| ((src[64 +4*y    +tabletx[tx]] & tablex[x]) ? 4 : 0)
						| ((src[64 +4*y +2 +tabletx[tx]] & tablex[x]) ? 8 : 0)
						;

					pix++;
				}
			}
		}
	}
};
